#include "net_topology.h"
#include <iostream>


/* constructor */

Net_Topology::Net_Topology(int sgsns_per_ggsn, int rncs_per_sgsn, int ues_per_rnc) {
	int i = 0;
	int j = 0;

	int num_of_sgsns = sgsns_per_ggsn;
	int num_of_rncs = num_of_sgsns*rncs_per_sgsn;
	int num_of_ues = ues_per_rnc*num_of_rncs;

	int RNC_id = 0;
	int SGSN_id = 2*num_of_rncs+num_of_ues;
 	int first_SGSN_id = SGSN_id;
	
	GGSN_id = first_SGSN_id + num_of_sgsns;

	net_list.clear();
	
	for (int UE_id = 2*num_of_rncs; UE_id < first_SGSN_id; UE_id++) {
		if (i == ues_per_rnc) {
			RNC_id++;
			i = 0;
		}
		if (j == rncs_per_sgsn*ues_per_rnc) {
			SGSN_id++;
			j = 0;
		}

		net_list.push_back(Net_Object(SGSN_id, RNC_id, UE_id));
		
		i++;
		j++;
 	}
}


/* access functions */

int Net_Topology::GetSize_of_list() const {
	return net_list.size();
}

int Net_Topology::Node_is_GGSN(int node_addr) {
	if (GGSN_id == node_addr)
		return 1;
	else
		return 0;
}

int Net_Topology::Node_is_SGSN(int node_addr) {
	list<Net_Object>::iterator element;
	for (element = net_list.begin() ; element != net_list.end() ; element++) {
		if (element->Get_SGSN_id() == node_addr) return 1;
	}
	return 0;
}

int Net_Topology::Node_is_RNC(int node_addr) {
	list<Net_Object>::iterator element;
	for (element = net_list.begin() ; element != net_list.end() ; element++) {
		if (element->Get_RNC_id() == node_addr) return 1;
	}
	return 0;
}

int Net_Topology::Node_is_UE(int node_addr) {
	list<Net_Object>::iterator element;
	for (element = net_list.begin() ; element != net_list.end() ; element++) {
		if (element->Get_UE_id() == node_addr) return 1;
	}
	return 0;
}

int Net_Topology::Get_SGSN_from_RNC(int node_addr) {
	if (!Node_is_RNC(node_addr)) return -1;
	
	list<Net_Object>::iterator element;
	for (element = net_list.begin() ; element != net_list.end() ; element++) {
		if (element->Get_RNC_id() == node_addr) return element->Get_SGSN_id();
	}
	return -1;
}

int Net_Topology::Get_SGSN_from_UE(int node_addr) {
	if (!Node_is_UE(node_addr)) return -1;
	
	list<Net_Object>::iterator element;
	for (element = net_list.begin() ; element != net_list.end() ; element++) {
		if (element->Get_UE_id() == node_addr) return element->Get_SGSN_id();
	}
	return -1;
}

int Net_Topology::Get_RNC_from_UE(int node_addr) {
	if( !Node_is_UE(node_addr)) return -1;
	
	list<Net_Object>::iterator element;
	for (element = net_list.begin() ; element != net_list.end() ; element++) {
		if (element->Get_UE_id() == node_addr) return element->Get_RNC_id();
	}
	return -1;
}

list<int> Net_Topology::Get_all_UEs() {
	list<int> UE_list;
	list<int> UE_unique_list;

	list<Net_Object>::iterator l_element;
	for (l_element = net_list.begin() ; l_element != net_list.end() ; l_element++) {
		UE_list.push_back(l_element->Get_UE_id());
	}
	UE_list.sort();
	unique_copy(UE_list.begin(), UE_list.end(), back_inserter(UE_unique_list));
	return UE_unique_list;
}

list<int> Net_Topology::Get_SGSNs_from_UEs(list<int> UE_list) {
	list<int> SGSN_list;
	list<int> SGSN_unique_list;

	list<int>::iterator UE_element;
	for (UE_element = UE_list.begin() ; UE_element != UE_list.end() ; UE_element++) {
		SGSN_list.push_back(this->Get_SGSN_from_UE(*UE_element));
	}
	SGSN_list.sort();
	unique_copy(SGSN_list.begin(), SGSN_list.end(), back_inserter(SGSN_unique_list));
	return SGSN_unique_list;
}

list<int> Net_Topology::Get_RNCs_from_UEs(list<int> UE_list, int SGSN_id) {
	list<int> RNC_list;
	list<int> RNC_unique_list;

	list<int>::iterator UE_element;
	for (UE_element = UE_list.begin() ; UE_element != UE_list.end() ; UE_element++) {
		if (this->Get_SGSN_from_UE(*UE_element) == SGSN_id)
			RNC_list.push_back(this->Get_RNC_from_UE(*UE_element));
	}
	RNC_list.sort();
	unique_copy(RNC_list.begin(), RNC_list.end(), back_inserter(RNC_unique_list));
	return RNC_unique_list;
}


list<int> Net_Topology::Get_UEs_under_RNC(list<int> UE_list, int RNC_id) {
	list<int> output_list;

	list<int>::iterator UE_element;
	for(UE_element = UE_list.begin() ; UE_element != UE_list.end() ; UE_element++) {
		if (this->Get_RNC_from_UE(*UE_element) == RNC_id)
			output_list.push_back(*UE_element);
	}
	return output_list;
}


/*list control functions */

void Net_Topology::clear_list() {
	net_list.clear();/*deletes all the elements from the list */
}

void Net_Topology::del_items_with_UE(int UE_id) {
	list<Net_Object>::iterator i;
	for(i = net_list.begin(); i != net_list.end() ; i++) {
		if (i->Get_UE_id() == UE_id) {
			net_list.erase(i);
			i--; /* Reduce the iterator because the next item took the place of current */
		}
	}
}

void Net_Topology::add_item(int SGSN_id, int RNC_id, int UE_id) {
	net_list.push_back(Net_Object(SGSN_id, RNC_id, UE_id));
}

void Net_Topology::print_all() {
	list<Net_Object>::iterator i;
	for(i = net_list.begin(); i != net_list.end() ; i++)
		printf(" %i\t%i\t%i\n", i->Get_SGSN_id(), i->Get_RNC_id(), i->Get_UE_id());
}

