/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the MASH Research
 * 	Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Header: /nfs/jade/vint/CVSROOT/ns-2/apps/rtp.cc,v 1.26 2000/08/18 18:34:01 haoboy Exp $";
#endif


#include <stdlib.h>

#include "config.h"
#include "agent.h"
#include "random.h"
#include "rtp.h"

/* File tcp.h was add because method recv must see struct hdr_tcp to get seqno. */
#include "tcp.h" 

//anpa>>
#define JOIN_GROUP            1
#define JOIN_GROUP_ACK        2
#define LEAVE_GROUP           3
#define LEAVE_GROUP_ACK       4
#define HANDOVER              5
#define HANDOVER_ACK          6
#define UPDATE_ACTIVE_SET     7
#define UPDATE_ACTIVE_SET_ACK 8
#define RELOCATION            9
#define RELOCATION_ACK       10
#define RELOCATION_FIN       11
//<<anpa

int hdr_rtp::offset_;

class RTPHeaderClass : public PacketHeaderClass {
public: 
        RTPHeaderClass() : PacketHeaderClass("PacketHeader/RTP",sizeof(hdr_rtp)) 
	{
		bind_offset(&hdr_rtp::offset_);
	}
} class_rtphdr;



static class RTPAgentClass : public TclClass {
public:
	RTPAgentClass() : TclClass("Agent/RTP") {}
	TclObject* create(int, const char*const*) {
		return (new RTPAgent());
	}
} class_rtp_agent;


RTPAgent::RTPAgent() : Agent(PT_RTP), session_(0), lastpkttime_(-1e6), 
    running_(0), rtp_timer_(this)
{
	bind("seqno_", &seqno_);
	bind_time("interval_", &interval_);
	bind("packetSize_", &size_);
	bind("maxpkts_", &maxpkts_);
	bind("random_", &random_);
//anpa>>
        bind("sgsns_per_ggsn_", &sgsns_per_ggsn_);
	bind("rncs_per_sgsn_", &rncs_per_sgsn_);
	bind("ues_per_rnc_", &ues_per_rnc_);
//<<anpa

}

void RTPAgent::start()
{ 
        running_ = 1;
        sendpkt();
        rtp_timer_.resched(interval_);
}


void RTPAgent::stop()
{   
        rtp_timer_.force_cancel();
        finish();
}

void RTPAgent::sendmsg(int nbytes, const char* /*flags*/)
{
        Packet *p;
        int n;

        if (++seqno_ < maxpkts_) {
                if (size_)
                        n = nbytes / size_;
                else
                        printf("Error: RTPAgent size = 0\n");

                if (nbytes == -1) {
                        start();
                        return;
                }
                while (n-- > 0) {
                        p = allocpkt();
                        hdr_rtp* rh = hdr_rtp::access(p);
                        rh->seqno() = seqno_;
                        target_->recv(p);
                }
                n = nbytes % size_;
                if (n > 0) {
                        p = allocpkt();
                        hdr_rtp* rh = hdr_rtp::access(p);
                        rh->seqno() = seqno_;
                        target_->recv(p);
                }
                idle();
        } else {
                finish();
                // xxx: should we deschedule the timer here? */
        };
}

void RTPAgent::timeout(int)
{
	if (running_) {
		sendpkt();
		
		if (session_)
			session_->localsrc_update(size_);
		double t = interval_;
		if (random_)
			/* add some zero-mean white noise */
			t += interval_ * Random::uniform(-0.5, 0.5);
		rtp_timer_.resched(t);
	}
}

/*
 * finish() is called when we must stop (either by request or because
 * we're out of packets to send.
 */
void RTPAgent::finish()
{
        running_ = 0;
        Tcl::instance().evalf("%s done", this->name());
}

void RTPAgent::advanceby(int delta)
{
        maxpkts_ += delta;
        if (seqno_ < maxpkts_ && !running_)
                start();
}               


void RTPAgent::recv(Packet* p, Handler*)
{
//anpa>>
// If the hole network topology has not been created yet, it must be created.
	if (net_topology.GetSize_of_list() == 0) Create_Net_Topology();

// Read the header of the packet.
	hdr_cmn* cmn = hdr_cmn::access(p);
	int packet_id = cmn->uid();
	hdr_rtp* pkt_header = hdr_rtp::access(p);
	
	if (pkt_header->packet_type() == JOIN_GROUP) {
		int starting_UE = pkt_header->starting_UE();
		join_routing(starting_UE, packet_id);
	} else if (pkt_header->packet_type() == JOIN_GROUP_ACK) {
		int starting_UE = pkt_header->starting_UE();
		join_ack_routing(starting_UE, packet_id);
	} else if (pkt_header->packet_type() == LEAVE_GROUP) {
		int starting_UE = pkt_header->starting_UE();
		leave_routing(starting_UE, packet_id);
	} else if (pkt_header->packet_type() == LEAVE_GROUP_ACK) {
		int starting_UE = pkt_header->starting_UE();
		leave_ack_routing(starting_UE, packet_id);
	} else if (pkt_header->packet_type() == HANDOVER) {
		int starting_UE = pkt_header->starting_UE();
		int new_addr = pkt_header->new_addr();
		int from_RNC = pkt_header->from_RNC();
		int to_RNC = pkt_header->to_RNC();
		handover_routing(starting_UE, new_addr, from_RNC, to_RNC, packet_id);
	} else if (pkt_header->packet_type() == HANDOVER_ACK) {
		int starting_UE = pkt_header->starting_UE();
		int new_addr = pkt_header->new_addr();
		int from_RNC = pkt_header->from_RNC();
		int to_RNC = pkt_header->to_RNC();
		handover_ack_routing(starting_UE, new_addr,  from_RNC, to_RNC, packet_id);
	} else if (pkt_header->packet_type() == UPDATE_ACTIVE_SET) {
		int starting_UE = pkt_header->starting_UE();
		int new_addr = pkt_header->new_addr();
		int from_RNC = pkt_header->from_RNC();
		int to_RNC = pkt_header->to_RNC();
		update_a_s_routing(starting_UE, new_addr,  from_RNC, to_RNC);
	} else if (pkt_header->packet_type() == UPDATE_ACTIVE_SET_ACK) {
		int starting_UE = pkt_header->starting_UE();
		int new_addr = pkt_header->new_addr();
		int from_RNC = pkt_header->from_RNC();
		int to_RNC = pkt_header->to_RNC();
		update_a_s_ack_routing(starting_UE, new_addr, from_RNC, to_RNC);
	} else if (pkt_header->packet_type() == RELOCATION) {
		int starting_UE = pkt_header->starting_UE();
		int new_addr = pkt_header->new_addr();
		int from_RNC = pkt_header->from_RNC();
		int to_RNC = pkt_header->to_RNC();
		relocation_routing(starting_UE, new_addr, from_RNC, to_RNC, packet_id);
	} else if (pkt_header->packet_type() == RELOCATION_ACK) {
		int starting_UE = pkt_header->starting_UE();
		int new_addr = pkt_header->new_addr();
		int from_RNC = pkt_header->from_RNC();
		int to_RNC = pkt_header->to_RNC();
		relocation_ack_routing(starting_UE, new_addr, from_RNC, to_RNC, packet_id);
	} else if (pkt_header->packet_type() == RELOCATION_FIN) {
		int starting_UE = pkt_header->starting_UE();
		int new_addr = pkt_header->new_addr();
		int from_RNC = pkt_header->from_RNC();
		int to_RNC = pkt_header->to_RNC();
		relocation_fin_routing(starting_UE, new_addr, from_RNC, to_RNC, packet_id);
	} else {
// Common packet for multicast routing has been received.
		common_routing(packet_id);
	}
//<<anpa
	if (session_)
	{
		session_->recv(p, 0);
	}
	else
	{
		Packet::free(p);
	}
}

int RTPAgent::command(int argc, const char*const* argv)
{
	if (argc == 2) {
		if (strcmp(argv[1], "rate-change") == 0) {
			rate_change();
			return (TCL_OK);
		} else if (strcmp(argv[1], "start") == 0) {
                        start();
			return (TCL_OK);
                } else if (strcmp(argv[1], "stop") == 0) {
                        stop();
                        return (TCL_OK);
                }
//anpa>>
		  else if (strcmp(argv[1], "join-mgroup") == 0) {
		  	join_mgroup();
		  	return (TCL_OK);
		} else if (strcmp(argv[1], "leave-mgroup") == 0) {
			leave_mgroup();
			return (TCL_OK);
		}
//<<anpa
	} else if (argc == 3) {
		if (strcmp(argv[1], "session") == 0) {
			session_ = (RTPSession*)TclObject::lookup(argv[2]);
			return (TCL_OK);
		} else if (strcmp(argv[1], "advance") == 0) {
                        int newseq = atoi(argv[2]);
                        advanceby(newseq - seqno_);
                        return (TCL_OK); 
                } else if (strcmp(argv[1], "advanceby") == 0) {
                        advanceby(atoi(argv[2]));
                        return (TCL_OK);
                }
//anpa>>
		  else if (strcmp(argv[1], "relocation") == 0) {
		  	relocation(atoi(argv[2]));
			return (TCL_OK);
		}
	}
	  else if (argc == 5) {
	  	if (strcmp(argv[1], "inter_RNS_handover") == 0) {
			inter_RNS_handover(atoi(argv[2]), atoi(argv[3]), here_.addr_, atoi(argv[4]));
			return (TCL_OK);
		}
	}
//<<anpa
	return (Agent::command(argc, argv));
}

/* 
 * We modify the rate in this way to get a faster reaction to the a rate
 * change since a rate change from a very low rate to a very fast rate may 
 * take an undesireably long time if we have to wait for timeout at the old
 * rate before we can send at the new (faster) rate.
 */
void RTPAgent::rate_change()
{
	rtp_timer_.force_cancel();
	
	double t = lastpkttime_ + interval_;
	
	double now = Scheduler::instance().clock();
	if ( t > now)
		rtp_timer_.resched(t - now);
	else {
		sendpkt();
		rtp_timer_.resched(interval_);
	}
}

void RTPAgent::sendpkt()
{
	Packet* p = allocpkt();
	lastpkttime_ = Scheduler::instance().clock();
	makepkt(p);
	target_->recv(p, (Handler*)0);
}

//anpa>>
void RTPAgent::sendpkt(int packet_id)
{
	Packet* p = allocpkt();
	hdr_cmn* cmn = hdr_cmn::access(p);
	cmn->uid() = packet_id;
	lastpkttime_ = Scheduler::instance().clock();
	makepkt(p);
	target_->recv(p, (Handler*)0);
}
//<<anpa

void RTPAgent::makepkt(Packet* p)
{
	hdr_rtp *rh = hdr_rtp::access(p);
	/* Fill in srcid_ and seqno */
	rh->seqno() = seqno_++;
	rh->srcid() = session_ ? session_->srcid() : 0;
}

void RTPTimer::expire(Event* /*e*/) {
        a_->timeout(0);
}

//anpa>>
//functions related with multicasting
void RTPAgent::join_mgroup()
{
// If the hole network topology has not been created yet, it must be created.
	if (net_topology.GetSize_of_list() == 0) Create_Net_Topology();

	dst_.addr_ = nsaddr_t(net_topology.Get_RNC_from_UE(here_.addr_));
	Packet* join_pkt = allocpkt();
	hdr_rtp* pkt_header = hdr_rtp::access(join_pkt);
	pkt_header->packet_type() = JOIN_GROUP;
	pkt_header->starting_UE() = here_.addr_;
	send(join_pkt, 0); 
}

void RTPAgent::leave_mgroup()
{
	dst_.addr_ = nsaddr_t(net_topology.Get_RNC_from_UE(here_.addr_));
	Packet* leave_pkt = allocpkt();
	hdr_rtp* pkt_header = hdr_rtp::access(leave_pkt);
	pkt_header->packet_type() = LEAVE_GROUP;
	pkt_header->starting_UE() = here_.addr_;
	send(leave_pkt, 0); 
}

void RTPAgent::inter_RNS_handover(int ue, int new_addr, int from_RNC, int to_RNC)
{
// If the hole network topology has not been created yet, it must be created.
	if (net_topology.GetSize_of_list() == 0) Create_Net_Topology();
	
	if (from_RNC != net_topology.Get_RNC_from_UE(ue)) {
		printf(" Sorry but UE: %i is not served by RNC: %i.\n", ue, from_RNC);
	} else {
		dst_.addr_ = nsaddr_t(to_RNC);
		Packet* ho_pkt = allocpkt();
		hdr_rtp* pkt_header = hdr_rtp::access(ho_pkt);
		pkt_header->packet_type() = HANDOVER;
		pkt_header->starting_UE() = ue;
		pkt_header->new_addr() = new_addr;
		pkt_header->from_RNC() = from_RNC;
		pkt_header->to_RNC() = to_RNC;
		Tcl::instance().evalf("$rnc(%i) add-gateway $rnc(%i)", from_RNC, to_RNC);
		send(ho_pkt, 0);
	}
}

void RTPAgent::relocation(int new_addr)
{
// Destination is SGSN over SRNC
	int serving_SGSN = net_topology.Get_SGSN_from_RNC(here_.addr_);
	dst_.addr_ = nsaddr_t(serving_SGSN);
// Create the packet
	Packet* ho_pkt = allocpkt();
	hdr_rtp* pkt_header = hdr_rtp::access(ho_pkt);
	pkt_header->packet_type() = RELOCATION;
// Find old address of relocated UE
	pkt_header->starting_UE() = Get_old_UE(new_addr);
	pkt_header->new_addr() = new_addr;
	pkt_header->from_RNC() = here_.addr_;
// Find DRNC
	pkt_header->to_RNC() = Get_DRNC(new_addr);
	Tcl::instance().evalf("$rnc(%i) add-gateway $sgsn([expr {%i - $num_of_rncs*(2 + $ues_per_rnc)}])",
			       here_.addr_, serving_SGSN);
	send(ho_pkt, 0);
}

void RTPAgent::common_routing(int packet_id)
{
	if (net_topology.Node_is_GGSN(here_.addr_)) {
		list<int> SGSN_list = net_topology.Get_SGSNs_from_UEs(multicast_group);
		for (list<int>::iterator i = SGSN_list.begin() ; i != SGSN_list.end() ; i++) {
			dst_.addr_ = nsaddr_t(*i);
			sendpkt(packet_id);
		}
	} else if (net_topology.Node_is_SGSN(here_.addr_)) {
		list<int> RNC_list = net_topology.Get_RNCs_from_UEs(multicast_group, here_.addr_);
		for (list<int>::iterator i = RNC_list.begin() ; i != RNC_list.end() ; i++) {
			dst_.addr_ = nsaddr_t(*i);
			sendpkt(packet_id);
		}
	} else if (net_topology.Node_is_RNC(here_.addr_)) {
		list<int> UE_list = net_topology.Get_UEs_under_RNC(multicast_group, here_.addr_);
		for (list<int>::iterator i = UE_list.begin() ; i != UE_list.end() ; i++) {
			dst_.addr_ = nsaddr_t(*i);
			sendpkt(packet_id);
		}
		list<int> RNC_list = Get_DRNCs(multicast_group);
		for (list<int>::iterator i = RNC_list.begin() ; i != RNC_list.end() ; i++) {
			dst_.addr_ = nsaddr_t(*i);
			Tcl::instance().evalf("$rnc(%i) add-gateway $rnc(%i)", here_.addr_, *i);
			sendpkt(packet_id);
		}
	} else if (net_topology.Node_is_UE(here_.addr_))
		/*printf("Received from UE: %i!\n", here_.addr_)*/;
	  else 
		printf("This node has no identity!\n");
}

void RTPAgent::join_routing(int starting_UE, int packet_id)
{
// If the hole network topology has not been created yet, it must be created.
	if (net_topology.GetSize_of_list() == 0) Create_Net_Topology();
	
	if (net_topology.Node_is_GGSN(here_.addr_)) {
		multicast_group.push_back(starting_UE);
		dst_.addr_ = nsaddr_t(net_topology.Get_SGSN_from_UE(starting_UE));
		Packet* join_next = allocpkt();
		hdr_rtp* pkt_header = hdr_rtp::access(join_next);
		pkt_header->packet_type() = JOIN_GROUP_ACK;
		pkt_header->starting_UE() = starting_UE;
		send(join_next, 0);
	} else if (net_topology.Node_is_SGSN(here_.addr_)) {
		dst_.addr_ = nsaddr_t(net_topology.Get_GGSN_id());
		Packet* join_next = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(join_next);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(join_next);
		pkt_header->packet_type() = JOIN_GROUP;
		pkt_header->starting_UE() = starting_UE;
		send(join_next, 0);
	} else if (net_topology.Node_is_RNC(here_.addr_)) {
		dst_.addr_ = nsaddr_t(net_topology.Get_SGSN_from_UE(starting_UE));
		Packet* join_next = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(join_next);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(join_next);
		pkt_header->packet_type() = JOIN_GROUP;
		pkt_header->starting_UE() = starting_UE;
		Tcl::instance().evalf("$rnc(%i) add-gateway $sgsn([expr {%i - $num_of_rncs*(2 + $ues_per_rnc)}])",
		                       here_.addr_, net_topology.Get_SGSN_from_RNC(here_.addr_));
		send(join_next, 0);
	} else if (net_topology.Node_is_UE(here_.addr_)) {
		/*printf("Received from UE: %i!\n", here_.addr_)*/;
	} else
		printf("This node has no identity!\n");
}

void RTPAgent::join_ack_routing(int starting_UE, int packet_id)
{
	if (net_topology.Node_is_SGSN(here_.addr_)) {
		multicast_group.push_back(starting_UE);
		dst_.addr_ = nsaddr_t(net_topology.Get_RNC_from_UE(starting_UE));
		Packet* join_next = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(join_next);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(join_next);
		pkt_header->packet_type() = JOIN_GROUP_ACK;
		pkt_header->starting_UE() = starting_UE;
		send(join_next, 0);
	} else if (net_topology.Node_is_RNC(here_.addr_)) {
		multicast_group.push_back(starting_UE);
		dst_.addr_ = nsaddr_t(starting_UE);
		Packet* join_next = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(join_next);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(join_next);
		pkt_header->packet_type() = JOIN_GROUP_ACK;
		pkt_header->starting_UE() = starting_UE;
		send(join_next, 0);
	} else if (net_topology.Node_is_UE(here_.addr_)) {
		;
	} else
		printf("This node has no identity!\n");
}

void RTPAgent::leave_routing(int starting_UE, int packet_id)
{
	if (net_topology.Node_is_GGSN(here_.addr_)) {
// Use an iterator to find the specific UE in the list.
		list<int>::iterator i;
		i = find(multicast_group.begin(), multicast_group.end() , starting_UE);
// Erase the element.
		multicast_group.erase(i);
		
		dst_.addr_ = nsaddr_t(net_topology.Get_SGSN_from_UE(starting_UE));
		Packet* leave_next = allocpkt();
		hdr_rtp* pkt_header = hdr_rtp::access(leave_next);
		pkt_header->packet_type() = LEAVE_GROUP_ACK;
		pkt_header->starting_UE() = starting_UE;
		send(leave_next, 0);
	} else if (net_topology.Node_is_SGSN(here_.addr_)) {
		dst_.addr_ = nsaddr_t(net_topology.Get_GGSN_id());
		Packet* leave_next = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(leave_next);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(leave_next);
		pkt_header->packet_type() = LEAVE_GROUP;
		pkt_header->starting_UE() = starting_UE;
		send(leave_next, 0);
	} else if (net_topology.Node_is_RNC(here_.addr_)) {
		dst_.addr_ = nsaddr_t(net_topology.Get_SGSN_from_UE(starting_UE));
		Packet* leave_next = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(leave_next);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(leave_next);
		pkt_header->packet_type() = LEAVE_GROUP;
		pkt_header->starting_UE() = starting_UE;
		Tcl::instance().evalf("$rnc(%i) add-gateway $sgsn([expr {%i - $num_of_rncs*(2 + $ues_per_rnc)}])",
		                       here_.addr_, net_topology.Get_SGSN_from_RNC(here_.addr_));
		send(leave_next, 0);
	} else if (net_topology.Node_is_UE(here_.addr_)) {
		/*printf("Received from UE: %i!\n", here_.addr_)*/;
	} else
		printf("This node has no identity!\n");
}

void RTPAgent::leave_ack_routing(int starting_UE, int packet_id)
{
	if (net_topology.Node_is_SGSN(here_.addr_)) {  
// Use an iterator to find the specific UE in the list.
		list<int>::iterator i;
		i = find(multicast_group.begin(), multicast_group.end(), starting_UE);
// Erase the element.
		multicast_group.erase(i);
		
		dst_.addr_ = nsaddr_t(net_topology.Get_RNC_from_UE(starting_UE));
		Packet* leave_next = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(leave_next);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(leave_next);
		pkt_header->packet_type() = LEAVE_GROUP_ACK;
		pkt_header->starting_UE() = starting_UE;
		send(leave_next, 0);
	} else if (net_topology.Node_is_RNC(here_.addr_)) {
// Use an iterator to find the specific UE in the list.
		list<int>::iterator i;
		i = find(multicast_group.begin(), multicast_group.end(), starting_UE);
// Erase the element.
		multicast_group.erase(i);
		dst_.addr_ = nsaddr_t(starting_UE);
		Packet* leave_next = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(leave_next);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(leave_next);
		pkt_header->packet_type() = LEAVE_GROUP_ACK;
		pkt_header->starting_UE() = starting_UE;
		send(leave_next, 0);
	} else if (net_topology.Node_is_UE(here_.addr_)) 
		;
	  else
		printf("This node has no identity!\n");
}

void RTPAgent::handover_routing(int starting_UE, int new_addr, int from_RNC, int to_RNC, int packet_id)
{
	if (net_topology.Node_is_RNC(here_.addr_)) {
// If the hole network topology has not been created yet, it must be created.
		if (net_topology.GetSize_of_list() == 0) Create_Net_Topology();
		
		dst_.addr_ = nsaddr_t(new_addr);
		Packet* ho_next = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(ho_next);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(ho_next);
		pkt_header->packet_type() = HANDOVER;
		pkt_header->starting_UE() = starting_UE;
		pkt_header->new_addr() = new_addr;
		pkt_header->from_RNC() = from_RNC;
		pkt_header->to_RNC() = to_RNC;
		Tcl::instance().evalf("$rnc(%i) add-gateway $rnc(%i)", to_RNC, from_RNC);
		send(ho_next, 0);
	} else {
// This case is executed in new node.
		net_topology.add_item(0, to_RNC, new_addr);
		dst_.addr_ = nsaddr_t(to_RNC);
		Packet* ho_ack = allocpkt();
		hdr_rtp* pkt_header = hdr_rtp::access(ho_ack);
		pkt_header->packet_type() = HANDOVER_ACK;
		pkt_header->starting_UE() = starting_UE;
		pkt_header->new_addr() = new_addr;
		pkt_header->from_RNC() = from_RNC;
		pkt_header->to_RNC() = to_RNC;
		send(ho_ack, 0);
	} 
}

void RTPAgent::handover_ack_routing(int starting_UE, int new_addr, int from_RNC, int to_RNC, int packet_id)
{
	if (here_.addr_ == to_RNC) {
		net_topology.add_item(net_topology.Get_SGSN_from_RNC(to_RNC), to_RNC, new_addr);
		multicast_group.push_back(new_addr);
		net_topology.del_items_with_UE(starting_UE);
		dst_.addr_ = nsaddr_t(from_RNC);
		Packet* ho_ack = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(ho_ack);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(ho_ack);
		pkt_header->packet_type() = HANDOVER_ACK;
		pkt_header->starting_UE() = starting_UE;
		pkt_header->new_addr() = new_addr;
		pkt_header->from_RNC() = from_RNC;
		pkt_header->to_RNC() = to_RNC;
		Tcl::instance().evalf("$rnc(%i) add-gateway $rnc(%i)", to_RNC, from_RNC);
		send(ho_ack, 0);
	} else if (here_.addr_ == from_RNC) {
		dst_.addr_ = nsaddr_t(starting_UE);
		Packet* ho_ack = allocpkt();
		hdr_rtp* pkt_header = hdr_rtp::access(ho_ack);
		pkt_header->packet_type() = UPDATE_ACTIVE_SET;
		pkt_header->starting_UE() = starting_UE;
		pkt_header->new_addr() = new_addr;
		pkt_header->from_RNC() = from_RNC;
		pkt_header->to_RNC() = to_RNC;
		send(ho_ack, 0); 
	} else
		printf("This packet should not have reached node: %i\n", here_.addr_);
}

void RTPAgent::update_a_s_routing(int starting_UE, int new_addr, int from_RNC, int to_RNC)
{
//This code is executed in starting UE
	dst_.addr_ = nsaddr_t(from_RNC);
	Packet* ho_ack = allocpkt();
	hdr_rtp* pkt_header = hdr_rtp::access(ho_ack);
	pkt_header->packet_type() = UPDATE_ACTIVE_SET_ACK;
	pkt_header->starting_UE() = starting_UE;
	pkt_header->new_addr() = new_addr;
	pkt_header->from_RNC() = from_RNC;
	pkt_header->to_RNC() = to_RNC;
	send(ho_ack, 0);
}

void RTPAgent::update_a_s_ack_routing(int starting_UE, int new_addr, int from_RNC, int to_RNC)
{
	if (net_topology.Node_is_RNC(here_.addr_)) {
		multicast_group.push_back(new_addr);
		net_topology.del_items_with_UE(starting_UE);
		net_topology.add_item(net_topology.Get_SGSN_from_RNC(to_RNC), to_RNC, new_addr);
		insert_in_drift_list(new_addr, to_RNC);
		insert_in_UE_pair_list(starting_UE, new_addr);
	} else
		printf("This packet should not have reached node: %i\n", here_.addr_);
}

void RTPAgent::relocation_routing(int starting_UE, int new_addr, int from_RNC, int to_RNC, int packet_id)
{
// If the hole network topology has not been created yet, it must be created.
	if (net_topology.GetSize_of_list() == 0) Create_Net_Topology();
	
	if ( here_.addr_ == net_topology.Get_SGSN_from_RNC(to_RNC)) {
// Agent in SGSN2 (step 2)
		dst_.addr_ = nsaddr_t(to_RNC);
// Create the packet
		Packet* reloc_pkt = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(reloc_pkt);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(reloc_pkt);
		pkt_header->packet_type() = RELOCATION;
		pkt_header->starting_UE() = starting_UE;
		pkt_header->new_addr() = new_addr;
		pkt_header->from_RNC() = from_RNC;
		pkt_header->to_RNC() = to_RNC;
		send(reloc_pkt, 0);
	} else if (here_.addr_ == net_topology.Get_SGSN_from_RNC(from_RNC)) {
// Agent in SGSN1 (step 1)
// Destination is SGSN over SRNC
		int to_SGSN = net_topology.Get_SGSN_from_RNC(to_RNC);
		dst_.addr_ = nsaddr_t(to_SGSN);
// Create the packet
		Packet* reloc_pkt = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(reloc_pkt);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(reloc_pkt);
		pkt_header->packet_type() = RELOCATION;
		pkt_header->starting_UE() = starting_UE;
		pkt_header->new_addr() = new_addr;
		pkt_header->from_RNC() = from_RNC;
		pkt_header->to_RNC() = to_RNC;
		send(reloc_pkt, 0);
	} else if ( here_.addr_ == to_RNC ) {
// Agent in RNC2 (step3)
		dst_.addr_ = nsaddr_t(from_RNC);
// Create the packet and send it to DRNC.
		Packet* reloc_pkt = allocpkt();
		hdr_rtp* pkt_header = hdr_rtp::access(reloc_pkt);
		pkt_header->packet_type() = RELOCATION_ACK;
		pkt_header->starting_UE() = starting_UE;
		pkt_header->new_addr() = new_addr;
		pkt_header->from_RNC() = from_RNC;
		pkt_header->to_RNC() = to_RNC;
		Tcl::instance().evalf("$rnc(%i) add-gateway $rnc(%i)", to_RNC, from_RNC);
		send(reloc_pkt, 0);
// Create the packet and send it to SGSN2.
		dst_.addr_ = nsaddr_t(net_topology.Get_SGSN_from_RNC(to_RNC));
		reloc_pkt = allocpkt();
		pkt_header = hdr_rtp::access(reloc_pkt);
		pkt_header->packet_type() = RELOCATION_ACK;
		pkt_header->starting_UE() = starting_UE;
		pkt_header->new_addr() = new_addr;
		pkt_header->from_RNC() = from_RNC;
		pkt_header->to_RNC() = to_RNC;
		Tcl::instance().evalf("$rnc(%i) add-gateway $sgsn([expr {%i - $num_of_rncs*(2 + $ues_per_rnc)}])",
		                       here_.addr_, net_topology.Get_SGSN_from_RNC(here_.addr_));
		send(reloc_pkt, 0);
	} else
		printf("This relocation packet should not have reach node: %i.\n", here_.addr_);
}

void RTPAgent::relocation_ack_routing(int starting_UE, int new_addr, int from_RNC, int to_RNC, int packet_id)
{
	if (here_.addr_ == from_RNC) {
// Agent in RNC1 (step 4)
		;
	} else if ( here_.addr_ == net_topology.Get_SGSN_from_RNC(to_RNC) ) {
// Agent in SGSN2 (step 5)
		dst_.addr_ = nsaddr_t(net_topology.Get_GGSN_id());
// Create the packet
		Packet* reloc_pkt = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(reloc_pkt);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(reloc_pkt);
		pkt_header->packet_type() = RELOCATION_ACK;
		pkt_header->starting_UE() = starting_UE;
		pkt_header->new_addr() = new_addr;
		pkt_header->from_RNC() = from_RNC;
		pkt_header->to_RNC() = to_RNC;
		send(reloc_pkt, 0);
	} else if ( here_.addr_ == net_topology.Get_GGSN_id() ) {
// Update Lists.
		list<int>::iterator i;
		i = find(multicast_group.begin(), multicast_group.end(), starting_UE);
		multicast_group.erase(i);
		multicast_group.push_back(new_addr);
		net_topology.del_items_with_UE(starting_UE);
		net_topology.add_item(net_topology.Get_SGSN_from_RNC(to_RNC), to_RNC, new_addr);
// Agent in GGSN (step6)
		dst_.addr_ = nsaddr_t(net_topology.Get_SGSN_from_RNC(from_RNC));
// Create the packet and send it to SGSN1.
		Packet* reloc_pkt = allocpkt();
		hdr_rtp* pkt_header = hdr_rtp::access(reloc_pkt);
		pkt_header->packet_type() = RELOCATION_FIN;
		pkt_header->starting_UE() = starting_UE;
		pkt_header->new_addr() = new_addr;
		pkt_header->from_RNC() = from_RNC;
		pkt_header->to_RNC() = to_RNC;
		send(reloc_pkt, 0);
		if ( net_topology.Get_SGSN_from_RNC(from_RNC) != net_topology.Get_SGSN_from_RNC(to_RNC) ) {
			dst_.addr_ = nsaddr_t(net_topology.Get_SGSN_from_RNC(to_RNC));
			reloc_pkt = allocpkt();
			pkt_header = hdr_rtp::access(reloc_pkt);
			pkt_header->packet_type() = RELOCATION_FIN;
			pkt_header->starting_UE() = starting_UE;
			pkt_header->new_addr() = new_addr;
			pkt_header->from_RNC() = from_RNC;
			pkt_header->to_RNC() = to_RNC;
			send(reloc_pkt, 0);
		}
	} else
		printf("This relocation packet ack should not have reach node: %i.\n", here_.addr_);
}

void RTPAgent::relocation_fin_routing(int starting_UE, int new_addr, int from_RNC, int to_RNC, int packet_id)
{
	if ( here_.addr_ == net_topology.Get_SGSN_from_RNC(from_RNC) ) {
// Agent in SGSN1 (step7).
// Update Lists.
		list<int>::iterator i;
		i = find(multicast_group.begin(), multicast_group.end(), starting_UE);
		multicast_group.erase(i);
		net_topology.del_items_with_UE(starting_UE);
		dst_.addr_ = nsaddr_t(from_RNC);
		Packet* reloc_pkt = allocpkt();
		hdr_cmn* cmn = hdr_cmn::access(reloc_pkt);
		cmn->uid() = packet_id;
		hdr_rtp* pkt_header = hdr_rtp::access(reloc_pkt);
		pkt_header->packet_type() = RELOCATION_FIN;
		pkt_header->starting_UE() = starting_UE;
		pkt_header->new_addr() = new_addr;
		pkt_header->from_RNC() = from_RNC;
		pkt_header->to_RNC() = to_RNC;
		send(reloc_pkt, 0);
		if ( net_topology.Get_SGSN_from_RNC(from_RNC) == net_topology.Get_SGSN_from_RNC(to_RNC) ) {
// Agent in SGSN2 (step8).
// Update Lists.
			multicast_group.push_back(new_addr);
			net_topology.add_item(net_topology.Get_SGSN_from_RNC(to_RNC), to_RNC, new_addr);
		}
	} else if ( here_.addr_ == net_topology.Get_SGSN_from_RNC(to_RNC) ) {
// Agent in SGSN2 (step8).
// Update Lists.
		multicast_group.push_back(new_addr);
		net_topology.del_items_with_UE(starting_UE);
		net_topology.add_item(net_topology.Get_SGSN_from_RNC(to_RNC), to_RNC, new_addr);
	} else if ( here_.addr_ == from_RNC ) {
// Agent in RNC2 (step9).
		remove_from_drift_list(new_addr);
		remove_from_UE_pair_list(new_addr);
	} else
		printf("This relocation fin packet should not have reach node: %i.\n", here_.addr_);
}

//access functions to net_topology
void RTPAgent::Create_Net_Topology()
{ net_topology = Net_Topology(sgsns_per_ggsn_, rncs_per_sgsn_, ues_per_rnc_); }

Net_Topology RTPAgent::Get_Net_Topology()
{ return net_topology; }

//access functions to drift list
void RTPAgent::insert_in_drift_list(int UE_id, int DRNC_id)
{
	drift_pair pair;
	pair.UE_id = UE_id;
	pair.DRNC_id = DRNC_id;
	drift_list.push_back(pair);
}

void RTPAgent::remove_from_drift_list(int UE_id)
{
	list<drift_pair>::iterator i;
	for(i = drift_list.begin(); i != drift_list.end() ; i++) {
		if (i->UE_id == UE_id) {
			drift_list.erase(i);
			i--; /* Reduce the iterator because the next item took the place of current */
		}
	}
}

int RTPAgent::Get_DRNC(int UE_id) 
{
	list<drift_pair>::iterator i;
	for(i = drift_list.begin(); i != drift_list.end() ; i++) {
		if (i->UE_id == UE_id) {
			return i->DRNC_id;
		}
	}
	return -1;
}

list<int> RTPAgent::Get_DRNCs(list<int> UE_list) {
	list<int> RNC_list;
	list<int> RNC_unique_list;
	list<int>::iterator UE_element;
	int DRNC;
	for (UE_element = UE_list.begin() ; UE_element != UE_list.end() ; UE_element++) {
		if ( (DRNC = Get_DRNC(*UE_element)) >= 0)
			RNC_list.push_back(DRNC);
	}
	RNC_list.sort();
	unique_copy(RNC_list.begin(), RNC_list.end(), back_inserter(RNC_unique_list));
	return RNC_unique_list;
}

// access functions to UE_pair list
void RTPAgent::insert_in_UE_pair_list(int old_UE, int new_UE) {
	UE_pair pair;
	pair.old_UE_id = old_UE;
	pair.new_UE_id = new_UE;
	UE_pair_list.push_back(pair);
}

void RTPAgent::remove_from_UE_pair_list(int new_UE_id)
{
	list<UE_pair>::iterator i;
	for(i = UE_pair_list.begin(); i != UE_pair_list.end() ; i++) {
		if (i->new_UE_id == new_UE_id) {
			UE_pair_list.erase(i);
			i--; /* Reduce the iterator because the next item took the place of current */
		}
	}
}

int RTPAgent::Get_new_UE(int old_UE) {
	list<UE_pair>::iterator i;
	for(i = UE_pair_list.begin(); i != UE_pair_list.end() ; i++) {
		if (i->old_UE_id == old_UE) {
			return i->new_UE_id;
		}
	}
	return -1;
}

int RTPAgent::Get_old_UE(int new_UE) {
	list<UE_pair>::iterator i;
	for(i = UE_pair_list.begin(); i != UE_pair_list.end() ; i++) {
		if (i->new_UE_id == new_UE) {
			return i->old_UE_id;
		}
	}
	return -1;
}
//<<anpa
