global ns

set ns [new Simulator]

set f [open test.tr w]
$ns trace-all $f

puts " Simulation started."

# Declare "finish" procedure which closes the trace file
proc finish {} {
	global ns f
	$ns flush-trace
	close $f
	puts " Simulation ended."
	exit 0
}

proc inter_RNS_handover { moving_ue serving_rnc drift_rnc } {
# Declare global variables
	global ns f
	global rnc bs ue
	global rtp_rnc rtp_ue psize
	global num_of_ues sgsns_per_ggsn rncs_per_sgsn

# Declare new UE.
	$ns node-config -UmtsNodeType ue \
			-baseStation $bs($drift_rnc) \
			-radioNetworkController $rnc($drift_rnc)

	set ue($num_of_ues) [$ns create-Umtsnode]

# Declare new agent on UE
	set rtp_ue($num_of_ues) [new Agent/RTP]
	$rtp_ue($num_of_ues) set fid_ 0
	$rtp_ue($num_of_ues) set packetSize_ $psize
	$rtp_ue($num_of_ues) set rate_ 6400kbps

# Attach the agent to the new UE and connect it to agent of RNC.
	$ns attach-agent $ue($num_of_ues) $rtp_ue($num_of_ues)
	$ns connect $rtp_rnc($drift_rnc) $rtp_ue($num_of_ues)
	set dch($num_of_ues) [$ns create-dch $ue($num_of_ues) $rtp_ue($num_of_ues)]

# Calculate the new node address of the moving UE.
	set new_ue_addr [expr {2*$sgsns_per_ggsn*$rncs_per_sgsn + + $sgsns_per_ggsn + $num_of_ues + 2}]

	$rtp_rnc($serving_rnc) inter_RNS_handover $moving_ue $new_ue_addr $drift_rnc
	$ue($num_of_ues) trace-inlink $f 2
	incr num_of_ues
}

######### DECLARATIONS #########
# Declare SGSN per GGSN ratio.
set sgsns_per_ggsn 2

# Declare RNC per GGSN ratio.
set rncs_per_sgsn 2

# Declare UE per RNC ratio.
set ues_per_rnc 4

# Declare packet size of rtp traffic
set psize 512

# Calculate the number of nodes.
set num_of_rncs [expr {$sgsns_per_ggsn * $rncs_per_sgsn}]
set num_of_ues [expr {$num_of_rncs * $ues_per_rnc}]

########### NODE CREATION ###########
# Create RNC's.
$ns node-config -UmtsNodeType rnc

for {set i 0} {$i < $num_of_rncs} {incr i} {
	set rnc($i) [$ns create-Umtsnode]
# Create Iur interfaces between RNC's.
	if { $i > 0} { 
		$ns duplex-link $rnc([expr {$i-1}]) $rnc($i) 622Mbit 0.4ms DropTail 1000
	}
}

# Create Node B's.
$ns node-config -UmtsNodeType bs \
		-downlinkBW 32kbs \
		-downlinkTTI 10ms \
		-uplinkBW 32kbs \
		-uplinkTTI 10ms 
		
for {set i 0} {$i < $num_of_rncs} {incr i} {
	set bs($i) [$ns create-Umtsnode]
# Create Iub interface between RNC's and Node B's.
	$ns setup-Iub $bs($i) $rnc($i) 622Mbit 622Mbit 15ms 15ms DummyDropTail 2000
}

# Create UE's
set j 0
set k 0

for {set i 0} {$i < $num_of_ues} {incr i} {
	if {$j == $ues_per_rnc} then {
		set j 0
	}
	if {$j == 0} then {
		$ns node-config -UmtsNodeType ue \
			-baseStation $bs($k) \
			-radioNetworkController $rnc($k)
		incr k
	}
	
	incr j
	set ue($i) [$ns create-Umtsnode]
}

# Create SGSN's.
set k 0

for {set i 0} {$i < $sgsns_per_ggsn} {incr i} {
	set sgsn($i) [$ns node]

	for {set j 0} {$j < $rncs_per_sgsn} {incr j} {
		$ns duplex-link $rnc($k) $sgsn($i) 622Mbit 0.4ms DropTail 1000
		incr k
	}
}

# Create GGSN.
set ggsn [$ns node]

for {set i 0} {$i < $sgsns_per_ggsn} {incr i} {
# Create Gn interface between GGSN and SGSN's.
	$ns duplex-link $sgsn($i) $ggsn 622MBit 10ms DropTail 1000
# Create Gn interface between SGSN's.
	if {$i != 0} then {
		$ns duplex-link $sgsn([expr {$i-1}]) $sgsn($i) 622Mbit 10ms DropTail 1000
	}
}

# Create external node.
set ext_node [$ns node]

$ns duplex-link $ggsn $ext_node 10MBit 35ms DropTail 1000

######### AGENT CREATION #########
# Create RTP agent in external node.
set rtp_ext [new Agent/RTP]
$rtp_ext set fid_ 0
$rtp_ext set packetSize_ $psize
$rtp_ext set rate_ 6400kbps
$rtp_ext set sgsns_per_ggsn_ $sgsns_per_ggsn
$rtp_ext set rncs_per_sgsn_ $rncs_per_sgsn
$rtp_ext set ues_per_rnc_ $ues_per_rnc

$ns attach-agent $ext_node $rtp_ext

# Create RTP agent in GGSN.
set rtp_ggsn [new Agent/RTP]
$rtp_ggsn set fid_ 0
$rtp_ggsn set packetSize_ $psize
$rtp_ggsn set rate_ 6400kbps
$rtp_ggsn set sgsns_per_ggsn_ $sgsns_per_ggsn
$rtp_ggsn set rncs_per_sgsn_ $rncs_per_sgsn
$rtp_ggsn set ues_per_rnc_ $ues_per_rnc

$ns attach-agent $ggsn $rtp_ggsn

$ns connect $rtp_ext $rtp_ggsn

# Create RTP agents in SGSNs.
for {set i 0} {$i < $sgsns_per_ggsn} {incr i} {
	set rtp_sgsn($i) [new Agent/RTP]
	$rtp_sgsn($i) set fid_ 0
	$rtp_sgsn($i) set packetSize_ $psize
	$rtp_sgsn($i) set rate_ 6400kbps
	$rtp_sgsn($i) set sgsns_per_ggsn_ $sgsns_per_ggsn
	$rtp_sgsn($i) set rncs_per_sgsn_ $rncs_per_sgsn
	$rtp_sgsn($i) set ues_per_rnc_ $ues_per_rnc
	
	$ns attach-agent $sgsn($i) $rtp_sgsn($i)

	$ns connect $rtp_ggsn $rtp_sgsn($i)
}

# Create RTP agents in RNCs.
set k 0;
set rtp_rnc($k) [new Agent/RTP]
for {set i 0} {$i < $sgsns_per_ggsn} {incr i} {
	for {set j 0} {$j < $rncs_per_sgsn} {incr j} {
		set rtp_rnc($k) [new Agent/RTP]
		$rtp_rnc($k) set fid_ 0
		$rtp_rnc($k) set packetSize_ $psize
		$rtp_rnc($k) set rate_ 6400kbps
		$rtp_rnc($k) set sgsns_per_ggsn_ $sgsns_per_ggsn
		$rtp_rnc($k) set rncs_per_sgsn_ $rncs_per_sgsn
		$rtp_rnc($k) set ues_per_rnc_ $ues_per_rnc
		
		$ns attach-agent $rnc($k) $rtp_rnc($k)
		
		$ns connect $rtp_sgsn($i) $rtp_rnc($k)

		incr k;
	}
}

# Connect RNC agents over the Iur interfaces.
for {set i 0} {$i < $num_of_rncs} {incr i} {
	if {$i > 0} { 
		$ns connect $rtp_rnc([expr {$i-1}]) $rtp_rnc($i)
	}
}

# Create RTP agents in UEs.
set k 0;

for {set i 0} {$i < $num_of_rncs} {incr i} {
	for {set j 0} {$j < $ues_per_rnc} {incr j} {
		set rtp_ue($k) [new Agent/RTP]
		$rtp_ue($k) set fid_ 0
		$rtp_ue($k) set packetSize_ $psize
		$rtp_ue($k) set rate_ 6400kbps
		$rtp_ue($k) set sgsns_per_ggsn_ $sgsns_per_ggsn
		$rtp_ue($k) set rncs_per_sgsn_ $rncs_per_sgsn
		$rtp_ue($k) set ues_per_rnc_ $ues_per_rnc
		
		$ns attach-agent $ue($k) $rtp_ue($k)
		
		$ns connect $rtp_rnc($i) $rtp_ue($k)

		incr k;
	}
}

######### TRAFFIC CREATION #########
# Read the video trace file:
set trace_file64 [new Tracefile]
$trace_file64 filename video64.dat

# Read the video trace file:
set video64 [new Application/Traffic/Trace]
$video64 attach-tracefile $trace_file64

$video64 attach-agent $rtp_ext

# Declare the session.
set session [new Session/RTP]
$session set uniq_srcid_ 5

$rtp_ext session $session
$rtp_ggsn session $session

# Declare RLC channels.
$ns node-config -llType UMTS/RLC/AM \
		-downlinkBW 384kbs \
		-uplinkBW 64kbs \
		-downlinkTTI 10ms \
		-uplinkTTI 20ms 

for {set i 0} {$i < $num_of_ues} {incr i} {
	set dch($i) [$ns create-dch $ue($i) $rtp_ue($i)]
}

# Trace UEs.
$ue(0) trace-inlink $f 2
$ue(4) trace-inlink $f 2

# Set the start/stop timing.
$ns at 0.0 "$video64 start"

$ns at 1.0 "$rtp_ue(0) join-mgroup"

$ns at 21.0 "inter_RNS_handover 8 0 1"
$ns at 41.0 "$rtp_rnc(0) relocation 28"

$ns at 59.0 "$rtp_ue(4) join-mgroup"
$ns at 61.0 "inter_RNS_handover 12 1 2"
$ns at 81.0 "$rtp_rnc(1) relocation 29" ;# New UE address must be given.
$ns at 101.9 "$video64 stop"
$ns at 102.0 "finish"
$ns run