##### NODE DECLARATIONS #####
set sgsns_per_ggsn 1
set rncs_per_sgsn 1
set ues_per_rnc 1
#############################

set interval 1.0
set duration 20.0
set psize 1000
set fid 0

set tcp_num 0
set pgmcc_num 0
set pgmcc_recv_num 0

ns-random 0

# better if bw and packet size don't match exactly (i.e., exactly one
# packet each 10ms), otherwise you might sometimes see strange
# synchronization effects in ns!
set bw 1007

Agent/TCP set packetSize_ $psize
Agent/PGM/Snd set packetSize_ $psize

# always add some randomness, s.o.
Agent/TCP set overhead_ 0.002

proc setupTCP {start stop s r} {
    global ns tcp_num fid interval tcp tcp_sink ftp
    set tcp($tcp_num) [new Agent/TCP/Sack1]
    $tcp($tcp_num) set fid_ $fid
    set tcp_sink($tcp_num) [new Agent/TCPSink/Sack1]
    $ns attach-agent $s $tcp($tcp_num)
    $ns attach-agent $r $tcp_sink($tcp_num)
    $ns connect $tcp($tcp_num) $tcp_sink($tcp_num)
    set ftp($tcp_num) [new Application/FTP]
    $ftp($tcp_num) attach-agent $tcp($tcp_num)
    $ns at $start "$ftp($tcp_num) start"
    $ns at $stop "$ftp($tcp_num) stop"
    incr tcp_num
    incr fid
    return $tcp([expr $tcp_num-1])
}

proc setupPGMCC {start stop s} {
    global ns pgmcc_num fid interval pgmcc
    set group [Node allocaddr]
    set pgmcc($pgmcc_num) [new Agent/PGM/Snd]
    $pgmcc($pgmcc_num) set fid_ $fid
    $pgmcc($pgmcc_num) set dst_addr_ $group
    $pgmcc($pgmcc_num) set dst_port_ 0
    $ns attach-agent $s $pgmcc($pgmcc_num)
    $ns at $start "$pgmcc($pgmcc_num) start"
    $ns at $stop "$pgmcc($pgmcc_num) stop"
    incr pgmcc_num
    incr fid
    return $pgmcc([expr $pgmcc_num-1])
}

proc setupPGMCCSink {start stop t r} {
    global ns pgmcc_recv_num pgmcc_sink
    set group [$t set dst_addr_]
    set pgmcc_sink($pgmcc_recv_num) [new Agent/PGM/Rcv]
    $ns attach-agent $r $pgmcc_sink($pgmcc_recv_num)
    $pgmcc_sink($pgmcc_recv_num) set dst_addr_ [$t set agent_addr_]
    $pgmcc_sink($pgmcc_recv_num) set dst_port_ [$t set agent_port_]
    $pgmcc_sink($pgmcc_recv_num) set recv_id_ $pgmcc_recv_num
    $ns at $start "$r join-group $pgmcc_sink($pgmcc_recv_num) $group"
    $ns at $stop "$pgmcc_sink($pgmcc_recv_num) stop; $r leave-group $pgmcc_sink($pgmcc_recv_num) $group"
    incr pgmcc_recv_num
    return $pgmcc_sink([expr $pgmcc_recv_num-1])
}

set ns [new Simulator -multicast on]

set f [open test.tr w]
$ns trace-all $f

########### NODE CREATION ###########
# Calculate the number of nodes.
set num_of_rncs [expr {$sgsns_per_ggsn * $rncs_per_sgsn}]
set num_of_ues [expr {$num_of_rncs * $ues_per_rnc}]

# Create RNC's.
for {set i 0} {$i < $num_of_rncs} {incr i} {
	set rnc($i) [$ns node]
}

# Create Node Bs and connect them to RNCs.
for {set i 0} {$i < $num_of_rncs} {incr i} {
	set bs($i) [$ns node]
	$ns duplex-link $bs($i) $rnc($i) 4Mbit 15ms DummyDropTail 200
}

# Create UEs and connect them to Node Bs.
set j 0
set k 0
for {set i 0} {$i < $num_of_ues} {incr i} {
	if {$j == $ues_per_rnc} then {
		set j 0
	}
	set ue($i) [$ns node]
	if {$j == 0} then {
		$ns simplex-link $bs($k) $ue($i) 384Kbit 15ms DummyDropTail 20
		$ns simplex-link $ue($i) $bs($k)  64Kbit 15ms DummyDropTail 10
		incr k
	}
	incr j
}

# Create SGSNs and connect them to RNCs.
set k 0
for {set i 0} {$i < $sgsns_per_ggsn} {incr i} {
	set sgsn($i) [$ns node]
	for {set j 0} {$j < $rncs_per_sgsn} {incr j} {
		$ns duplex-link $rnc($k) $sgsn($i) 4Mbit 0.4ms DropTail 100
		incr k
	}
}

# Create GGSN and connect it to SGSNs.
set ggsn [$ns node]
for {set i 0} {$i < $sgsns_per_ggsn} {incr i} {
	$ns duplex-link $sgsn($i) $ggsn 4MBit 10ms DropTail 100
}

# Create external node and connect it to GGSN.
set ext_node [$ns node]
$ns duplex-link $ggsn $ext_node 1MBit 35ms DropTail 100

set mproto DM
set mrthandle [$ns mrtproto $mproto {}]

set t [setupPGMCC 1.0 $duration $ext_node]
for {set i 0} {$i < $num_of_ues} {incr i} {
    setupPGMCCSink 0.9 $duration $t $ue($i)
}
#for {set i 0} {$i < $num_of_ues} {incr i} {
#    setupTCP 2.0 $duration $ext_node $ue($i)
#}

$ns at $duration "exit 0"
$ns run
