/*
 * pgm.cc - 000330
 * Copyright (c) 2000 Gianluca Iannaccone, Luigi Rizzo
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>

#include "tclcl.h"
#include "agent.h"
#include "my_pgm.h"

int hdr_pgm::offset_;
int hdr_pgm_spm::offset_;
int hdr_pgm_odata::offset_;
int hdr_pgm_rdata::offset_;
int hdr_pgm_nak::offset_;
int hdr_pgm_nnak::offset_;
int hdr_pgm_ncf::offset_;
int hdr_pgm_ack::offset_ ;

static class PGMAgentClass : public TclClass {
public:
  PGMAgentClass() : TclClass("Agent/PGM") {}
  TclObject* create(int, const char*const*) {
    return (new PGMAgent());
  }
} class_pgm_agent;

static class PGMHeaderClass : public PacketHeaderClass {
public:
  PGMHeaderClass() : PacketHeaderClass("PacketHeader/PGM",
				       sizeof(hdr_pgm)) {
	bind_offset(&hdr_pgm::offset_); 
  }
} class_pgmhdr;

static class PGM_SPMHeaderClass : public PacketHeaderClass {
public:
  PGM_SPMHeaderClass() : PacketHeaderClass("PacketHeader/PGM_SPM",
					   sizeof(hdr_pgm_spm)) {
	bind_offset(&hdr_pgm_spm::offset_); 
  }
} class_pgm_spmhdr;

static class PGM_OdataHeaderClass : public PacketHeaderClass {
public:
  PGM_OdataHeaderClass() : PacketHeaderClass("PacketHeader/PGM_Odata",
					     sizeof(hdr_pgm_odata)) {
	bind_offset(&hdr_pgm_odata::offset_); 
  }
} class_pgm_odatahdr;

static class PGM_RdataHeaderClass : public PacketHeaderClass {
public:
  PGM_RdataHeaderClass() : PacketHeaderClass("PacketHeader/PGM_Rdata",
					     sizeof(hdr_pgm_rdata)) {
	bind_offset(&hdr_pgm_rdata::offset_); 
  }
} class_pgm_rdatahdr;

static class PGM_NakHeaderClass : public PacketHeaderClass {
public:
  PGM_NakHeaderClass() : PacketHeaderClass("PacketHeader/PGM_Nak",
					   sizeof(hdr_pgm_nak)) {
	bind_offset(&hdr_pgm_nak::offset_); 
  }
} class_pgm_nakhdr;

static class PGM_NNakHeaderClass : public PacketHeaderClass {
public:
  PGM_NNakHeaderClass() : PacketHeaderClass("PacketHeader/PGM_NNak",
					   sizeof(hdr_pgm_nnak)) {
	bind_offset(&hdr_pgm_nnak::offset_); 
  }
} class_pgm_nnakhdr;

static class PGM_NcfHeaderClass : public PacketHeaderClass {
public:
  PGM_NcfHeaderClass() : PacketHeaderClass("PacketHeader/PGM_Ncf",
					   sizeof(hdr_pgm_ncf)) {
	bind_offset(&hdr_pgm_ncf::offset_); 
  }
} class_pgm_ncfhdr;

static class PGM_AckHeaderClass : public PacketHeaderClass {
public:
  PGM_AckHeaderClass() : PacketHeaderClass("PacketHeader/PGM_Ack",
                                           sizeof(hdr_pgm_ack)) {
        bind_offset(&hdr_pgm_ack::offset_); }
} class_pgm_ackhdr;

// Base PGM Agent

PGMAgent::PGMAgent() : 
Agent(PT_PGM), 
rng_(RNG::RAW_SEED_SOURCE),
xplotchannel_(0),
pkts_sent_(0),
pkts_rcvd_(0),
odata_sent_(0),     
odata_rcvd_(0),
rdata_sent_(0),
rdata_rcvd_(0),
spm_sent_(0),     
spm_rcvd_(0),
nak_sent_(0),     
nak_rcvd_(0),
ncf_sent_(0),
ncf_rcvd_(0),
nnak_sent_(0),
nnak_rcvd_(0),
ack_sent_(0),
ack_rcvd_(0),
unrecoverable_pkts_(0) 
{
  bind("off_pgm_", &off_pgm_);
  bind("off_pgm_spm_", &off_pgm_spm_);
  bind("off_pgm_odata_", &off_pgm_odata_);
  bind("off_pgm_rdata_", &off_pgm_rdata_);
  bind("off_pgm_nak_", &off_pgm_nak_);
  bind("off_pgm_nnak_", &off_pgm_nnak_);
  bind("off_pgm_ncf_", &off_pgm_ncf_);
  bind("off_pgm_ack_", &off_pgm_ack_);

  bind("color_", &color_) ;
}

int 
PGMAgent::command(int argc, const char*const* argv)
{
  Tcl& tcl = Tcl::instance();

  if ( argc == 2 ) 
    if ( !strcmp(argv[1], "print-statistics") ) {
      print_statistics();
      return TCL_OK;
    } else if ( !strcmp(argv[1], "flush") ) {
      if ( xplotchannel_ ) 
         Tcl_Flush(xplotchannel_) ; ;
      return TCL_OK ;
    } else if ( !strcmp(argv[1], "clear-statistics") ) {
      pkts_sent_ = pkts_rcvd_ = 0 ;
      odata_sent_ = odata_rcvd_ = 0 ;
      rdata_sent_ = rdata_rcvd_ = 0 ;
      spm_sent_ = spm_rcvd_ = 0 ;
      nak_sent_ = nak_rcvd_ = 0 ;
      ncf_sent_ = ncf_rcvd_ = 0 ;
      nnak_sent_ = nnak_rcvd_ = 0 ;
      ack_sent_ = ack_rcvd_ = 0 ;
      unrecoverable_pkts_ = 0 ;
	  return TCL_OK ;
	} 

  if ( argc == 3 ) 
    if ( !strcmp(argv[1], "xfile-attach") ) {
       int mode;
       char* id = (char *) argv[2];
       xplotchannel_ = Tcl_GetChannel(tcl.interp(), id, &mode);
       if (xplotchannel_ == 0) {
           tcl.resultf("trace: can't attach %s for writing", id);
           return (TCL_ERROR);
       }
       return (TCL_OK);
    } else if ( !strcmp(argv[1], "results") ) {
		printf("+ %d %d %d\n", atoi(argv[2]), odata_sent_ + rdata_sent_, color_) ;
		fflush(stdout) ;
		return (TCL_OK) ;
    }

   return Agent::command(argc, argv);
}

void
PGMAgent::print_statistics() 
{
   printf("Packets:\t%5d sent, %5d rcvd\n"
          "ODATA:\t%5d sent, %5d rcvd\n"
          "RDATA:\t%5d sent, %5d rcvd\n"
          "SPM:\t%5d sent, %5d rcvd\n"
          "NAK:\t%5d sent, %5d rcvd\n"
          "NCF:\t%5d sent, %5d rcvd\n"
          "NNAK:\t%5d sent, %5d rcvd\n"
          "ACK:\t%5d sent, %5d rcvd\n"
          "Unrecoverable pkts: %d\n",
          pkts_sent_, pkts_rcvd_,
          odata_sent_, odata_rcvd_,
          rdata_sent_, rdata_rcvd_,
          spm_sent_, spm_rcvd_,
          nak_sent_, nak_rcvd_,
          ncf_sent_, ncf_rcvd_,
          nnak_sent_, nnak_rcvd_,
          ack_sent_, ack_rcvd_,
          unrecoverable_pkts_) ;
}

void
PGMAgent::print_xplot(Packet * p) 
{
   if ( !xplotchannel_ )
      return ;
//correction-anpa-   hdr_cmn * ch = (hdr_cmn *) p->access(off_cmn_) ;
//correction-anpa>>
   	hdr_cmn* ch = hdr_cmn::access(p);
//<<correction-anpa
      
      

   hdr_pgm * ph = (hdr_pgm *) p->access(off_pgm_) ;
   double now = Scheduler::instance().clock() ;
   u_int seqno = ph->sqn() ;
   char xstr[50] ;

   switch ( ch->ptype() ) {
   case PT_PGM_ODATA:
      sprintf(xstr, "+ %f %u %d\n", now, seqno, color_) ;
      break ;
   case PT_PGM_RDATA:
      sprintf(xstr, "+ %f %u yellow\n", now, seqno) ;
      break ;
   case PT_PGM_NAK:
      sprintf(xstr, "box %f %u white\n", now, seqno) ;
      break ;
   case PT_PGM_NCF:
      sprintf(xstr, "diamond %f %u white\n", now, seqno) ;
      break ;
   default:
      return ; 
   }

   Tcl_Write(xplotchannel_, xstr, strlen(xstr));
}