/*
 * pgmcc.cc - 000614
 * Copyright (c) 2000 Gianluca Iannaccone, Luigi Rizzo
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
* OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *      
 */

#include <stdio.h>
#include <math.h>
#include <tcl.h>

#include "template.h"
#include "agent.h"
#include "packet.h"
#include "queue.h"
#include "my_pgmcc.h"

static class PGMCCSndAgentClass : public TclClass {
public:
  	PGMCCSndAgentClass() : TclClass("Agent/PGM/Snd/CC") {}
  	TclObject* create(int, const char*const*) {
    	return (new PGMCCSndAgent());
  	}
} class_pgm_snd_cc_agent;

// PGMCC Sender Agent

PGMCCSndAgent::PGMCCSndAgent():
xplotlosschannel_(0),
xplotpktlosschannel_(0),
cc_timer_(this, CC_TIMEOUT),
cc_token_(1),
cc_window_(1),
dupacks_(0),
ack_lead_(0),
ignore_cong_(0),
ack_bitmask_(~0),
acker_mrtt_(0),
acker_loss_(0),
acker_addr_(0),
session_startup_(1) 
{
	bind("ss_thresh_",&ss_thresh_);
  	bind("dup_thresh_",&dup_thresh_);
  	bind("weight_", &weight_);
  	bind("cc_timeout_",&cc_timeout_);
  	bind("hyst_",&hyst);
  	bind_bool("use_timestamp_", &use_timestamp_) ;
  	bind_bool("use_ewma_rtt_", &use_ewma_rtt_) ;
  	bind_bool("use_adaptive_ss_", &use_adaptive_ss_) ;
}

void
PGMCCSndAgent::start(void)
{
    cc_timer_.resched(cc_timeout_);
  	PGMSndAgent::start() ;
}

int
PGMCCSndAgent::command(int argc, const char*const* argv)
{
  	Tcl& tcl = Tcl::instance();
 
  	if ( argc == 3 )
    	if ( !strcmp(argv[1], "cc-report") ) {
       		int mode;
       		char* id = (char *) argv[2];
       		xplotlosschannel_ = Tcl_GetChannel(tcl.interp(), id, &mode);
       		if (xplotlosschannel_ == 0) {
           		tcl.resultf("trace: can't attach %s for writing", id);
           		return (TCL_ERROR);
       		} 
       	return (TCL_OK);
   	} else if ( !strcmp(argv[1], "acker-report") ) {
    	int mode;
       	char* id = (char *) argv[2];
       	xplotpktlosschannel_ = Tcl_GetChannel(tcl.interp(), id, &mode);
       	if (xplotpktlosschannel_ == 0) {
           	tcl.resultf("trace: can't attach %s for writing", id);
           	return (TCL_ERROR);
       	}
       	return (TCL_OK);  
    }

   	return PGMSndAgent::command(argc, argv);
}

void 
PGMCCSndAgent::recv(Packet * p, Handler * h)
{
//correction-anpa-   hdr_cmn * ch = (hdr_cmn *) p->access(off_cmn_) ;
//correction-anpa>>
   	hdr_cmn* ch = hdr_cmn::access(p);
//<<correction-anpa
 	hdr_pgm *ph = (hdr_pgm *) p->access(off_pgm_);

  	if ( ch->ptype() != PT_PGM_ACK ) {
     	PGMSndAgent::recv(p, h) ;
     	return ;
  	}

  	if ( ph->options() & PGM_OPT_PRESENT ) 
    	option_handler(p) ;

  	print_xplot(p) ;

  	// It's an ack. Reset the the cc_timeout 
  	cc_timer_.resched(cc_timeout_);	

  	hdr_pgm_ack *pah = (hdr_pgm_ack *) p->access(off_pgm_ack_);
  	u_int lead = pah->ack_rxw_lead() ;
  	u_int32_t bitmask = pah->bitmask();
  	if ( pah->nacker() == acker_addr_ ) {
      	double rtt = use_timestamp_ ? 
                     Scheduler::instance().clock() - pah->timestamp() :
                     txw_lead_ - lead ;
      	acker_mrtt_ = use_ewma_rtt_ ?
				   	  acker_mrtt_ * weight_  + rtt * ( 1 - weight_ ) :
				      rtt ; 
      	acker_loss_ = ( acker_mrtt_ * acker_mrtt_ ) * pah->rx_loss() ;
#ifdef LOSS
		fprintf(stderr, "+ %f %f red\nx %f %f yellow\n",
				Scheduler::instance().clock(), acker_mrtt_,
				Scheduler::instance().clock(), pah->rx_loss()) ;
#endif
      	if (xplotlosschannel_) {
          	char xstr[50];
          	sprintf(xstr, "box %u %f %d\n", lead, acker_mrtt_, color_) ;
          	Tcl_Write(xplotlosschannel_, xstr, strlen(xstr));
          	sprintf(xstr, "diamond %u %f green\n", lead, acker_loss_) ;
          	Tcl_Write(xplotlosschannel_, xstr, strlen(xstr));
      	}
	  	if (xplotpktlosschannel_) {
          	char xstr[50];
          	sprintf(xstr, "x %u %f %d\n", lead, pah->rx_loss(), color_) ;
          	Tcl_Write(xplotpktlosschannel_, xstr, strlen(xstr));
	  	}
  	}

 	/*
   	 * using bitmasks, determine new acks carried by this pkt.
   	 * ack_bitmask_ is the old state, bitmask is the one in the ack.
  	 * This section is necessary because we dont have cumulative acks
   	 */
  	int delta = (int) (lead - ack_lead_) ;
  	if (delta > 32) // very new, reset old state 
     	ack_bitmask_ = 0 ;
  	else if (delta > 0)
     	ack_bitmask_ <<= delta ;
  	else if (delta < -32)
     	bitmask = 0 ;
  	else
     	bitmask <<= -delta ;

  	// compute the new acks 
  	u_int32_t new_acks = bitmask & ~ack_bitmask_ ;
  	// and update state 
  	ack_bitmask_ |= bitmask ;
  	if ( pgm_seq_gt(lead, ack_lead_) )
     	ack_lead_ = lead ;

  	// compute total acks in this packet
 	int total, missing ;
  	u_int32_t l; 
  	for (total = 0, l = new_acks ; l ; l >>= 1 )
     	if (l & 1)
        	total++ ;

  	// count missing packets 
  	for (missing = 0 , l = ~ack_bitmask_ ; l ; l >>= 1)
     	if (l & 1)
        	missing++ ;

  	if (total == 0) {
     	// fully duplicate acks 
     	Packet::free(p) ;
     	return ;
  	}

  	// Do not count dups for ACKs within an rtt from prev. cong.
  	if ( pgm_seq_leq(lead, ignore_cong_) ) {
     	missing = 0 ;
     	ack_bitmask_ = ~0 ;
  	}
 
  	if ( missing == 0 ) {
    	total += dupacks_ ; // recover previous acks 
     	dupacks_ = 0 ;
     	double delta_w ;
     	if (cc_token_ < 0)
         	delta_w = 0 ;	     // do not increase window
     	else if (cc_window_ < ss_thresh_)
         	delta_w = 1 ;       // exp. increase for small windows 
     	else
         	delta_w = 1. / cc_window_ ;
     	cc_token_ += total * (1 + delta_w) ;
     	cc_window_ += total * delta_w ;  
#ifdef PGMCC
     	fprintf(stderr, "-- [%5.3f] window_size: %f tokens: %f\n",
             	Scheduler::instance().clock(), cc_window_, cc_token_) ;
#endif
  	} else {
     	dupacks_ += total ;
     	if (dupacks_ >= dup_thresh_) {
         	/*
          	 * re-sync window estimate with reality.
	 	 	 * Since we know the real window, we also know how
	  		 * many tokens will arrive (if no loss), and how
	  		 * many we should ignore.
	  	 	 */
         	cc_window_ = (txw_lead_ - lead + 1 );
			// mimic TCP slowstart 
			if ( use_adaptive_ss_ ) 
				ss_thresh_ = cc_window_ / 2 ;
	 		ignore_cong_ = txw_lead_ ;
	 		if (cc_token_ < 0) 
	    		cc_token_++ ;
	 		else if (cc_window_ < 4 )  // small window
            	cc_token_ = 1 ;
         	else {
	    		cc_window_ /= 2 ;
            	cc_token_ = -cc_window_ + 1;
 		} 
	 		ack_bitmask_ = ~0 ; // not count anymore these losses 
	 		dupacks_ = 0 ;
#ifdef PGMCC
         	fprintf(stderr, 
					"-- [%5.3f] congestion. window_size: %f tokens: %f\n",
                 	Scheduler::instance().clock(), cc_window_, cc_token_) ;
#endif
      	} else if ( txw_lead_ == lead ) 
			/* 
			 * XXX if there are no flying-packets we forcely send a new 
			 * 	   one to avoid a stall
			 */
			cc_token_ = 1 ; 
  	}  
#ifdef PGMCC
  	fprintf(stderr, "-- [%5.3f] ACK (%d:%d): seq %u mask %x loss %f ts %f\n",
          	Scheduler::instance().clock(), 
		  	pah->nacker() >> 8 , pah->nacker() & 0xff, ph->sqn(), 
          	pah->bitmask(), pah->rx_loss(), pah->timestamp()) ;
#endif
  	Packet::free(p) ; // don't need this packet anymore ;
  	data_move();
}

int 
PGMCCSndAgent::can_tx() 
{
	if ( PGMSndAgent::can_tx() && cc_token_ >= 1 ) {
		cc_token_-- ;
		return 1 ;
	}
	return 0 ;
}

void
PGMCCSndAgent::next_packet() 
{
    if ( cc_token_ < 1 ) {
		next_packet_timer_.force_cancel() ;
		return ;
	}   
	PGMSndAgent::next_packet() ;
}

void
PGMCCSndAgent::timeout(int tno)
{
   	if ( tno != CC_TIMEOUT ) {
      	PGMSndAgent::timeout(tno) ;
      	return;
   	}
   	ignore_cong_ = txw_lead_;
   	ack_lead_ = txw_lead_ ;
   	cc_token_ = 1;
   	cc_window_ = 1;
   	dupacks_ = 0;
   	ack_bitmask_ = ~0;
   	// XXX - do not elect a new acker 
   	// xmit_options_ |= PGM_OPT_SEND_NAK;
   	// acker_addr_ = 0;
   	// if ( use_adaptive_ss_ && session_startup_ ) {
	// 		session_startup_ = 0 ;
	// 		ss_thresh_ = sendspace_ ;
	// 	} 
   	cc_timer_.resched(cc_timeout_);
#ifdef PGMCC
   	fprintf(stderr, "-- [%5.3f] window_size: %f tokens: %f\n",
   			Scheduler::instance().clock(), cc_window_, cc_token_) ;
#endif
   	data_move() ;
}

void
PGMCCSndAgent::option_handler(Packet *p)
{
  	hdr_pgm *ph = (hdr_pgm *) p->access(off_pgm_);

  	if ( !ph->opttype() ) 
     	return ;

  	if ( ph->opttype() & PGM_OPT_LOSSRATE ) {
     	ph->opttype() &= ~PGM_OPT_LOSSRATE ;
     	hdr_pgm_nak * pnh = (hdr_pgm_nak *) p->access(off_pgm_nak_);
     	double nacker_loss ;
     	int acker_switch = 0 ; 
     	double rttn = use_timestamp_ ?
                   		Scheduler::instance().clock() - pnh->timestamp() :
                   		txw_lead_ - pnh->rx_lead() ;  
		// XXX adjustement to rtt estimate when using sequence number
		if ( !use_timestamp_ ) {
			rttn -= 1 + ( ss_thresh_ >= (int) cc_window_ ) ;
			if ( rttn < 0 ) 
				rttn = 0 ;
		}
		nacker_loss = rttn * rttn * pnh->rx_loss() ; 

#ifdef PGMCC
     	fprintf(stderr, 
				"-- [%5.3f] NAK (%d:%d): seq %d rxlead %d loss %f ts %f\n",
             	Scheduler::instance().clock(), 
		     	pnh->nacker() >> 8 , pnh->nacker() & 0xff, 
             	ph->sqn(), pnh->rx_lead(), pnh->rx_loss(), pnh->timestamp() ) ;
#endif

     	if ( !acker_addr_ ) {
			acker_addr_ = pnh->nacker() ;
			acker_mrtt_ = rttn ;
			cc_token_++ ; // consider this as an ACK 
			ack_lead_ = txw_lead_ ;
			ack_bitmask_ = ~0 ;
        	data_move() ;
        	acker_switch = 1 ;
#ifdef PGMCC
        	fprintf(stderr, "-- [%5.3f] new election: acker %d:%d\n",
                	Scheduler::instance().clock(), 
			 		acker_addr_ >> 8 , acker_addr_ & 0xff ) ;
#endif
     	} else if ( nacker_loss > (acker_loss_ * hyst) && 
					pnh->nacker() != acker_addr_) {
#ifdef PGMCC
        	fprintf(stderr, "-- [%5.3f] acker switch (%d:%d -> %d:%d)\n", 
                	Scheduler::instance().clock(), 
					acker_addr_ >> 8, acker_addr_ & 0xff, 
					pnh->nacker() >> 8, pnh->nacker() & 0xff ) ;
        	fprintf(stderr, "             old loss %f, new loss %f\n"
                	"             old rtt %f, new rtt %f\n",
                	acker_loss_, nacker_loss, acker_mrtt_, rttn) ;
#endif
			acker_addr_ = pnh->nacker() ;
			acker_mrtt_ = rttn ;
#ifdef LOSS 
 			double now = Scheduler::instance().clock() ;
			fprintf(stderr, 
					"line %f 0 %f 50 green\n"
					"+ %f %f red\n"
					"x %f %f yellow\n",
					now, now, now, acker_mrtt_, now, pnh->rx_loss()) ;
#endif
        	acker_switch = 1 ;
     	}

     	if ( acker_switch ) {
        	char xstr[100] ;

     		if ( xplotchannel_ ) {
        		double now = Scheduler::instance().clock() ;
        		sprintf(xstr, 
						"line %f 0 %f %u green\n", 
						now, now, pnh->rx_lead()) ;
        		Tcl_Write(xplotchannel_, xstr, strlen(xstr));
     		}

     		if ( xplotlosschannel_ ) { 
        		sprintf(xstr, 
						"line %u 0 %u %f %d\n", 
                 		pnh->rx_lead(), pnh->rx_lead(), acker_mrtt_, color_) ;
        		Tcl_Write(xplotlosschannel_, xstr, strlen(xstr));
     		}

	 		if ( xplotpktlosschannel_ ) {
        		sprintf(xstr, 
						"line %u 0 %u %f %d\n", 
            			pnh->rx_lead(), pnh->rx_lead(), acker_mrtt_, color_) ;
        		Tcl_Write(xplotpktlosschannel_, xstr, strlen(xstr));
	 		}
	 	}

     	if ( acker_addr_ == pnh->nacker() )  
			acker_loss_ = ( (acker_mrtt_ * acker_mrtt_) * pnh->rx_loss()); 

     	if (xplotlosschannel_ ) { 
       		char xstr[50];
       		sprintf(xstr, "box %u %f red\n", pnh->rx_lead(), rttn) ;
       		Tcl_Write(xplotlosschannel_, xstr, strlen(xstr));
       		sprintf(xstr, "diamond %u %f red\n", pnh->rx_lead(), nacker_loss) ;
       		Tcl_Write(xplotlosschannel_, xstr, strlen(xstr));
     	}
	 	if ( xplotpktlosschannel_ ) {
       		char xstr[50];
       		sprintf(xstr, "x %u %f red\n", pnh->rx_lead(), pnh->rx_loss()) ;
      		Tcl_Write(xplotpktlosschannel_, xstr, strlen(xstr));
	 	}
  	}

  	PGMSndAgent::option_handler(p) ;
}

void
PGMCCSndAgent::buildoptions(Packet *p, packet_t type)
{
//correction-anpa-   hdr_cmn * ch = (hdr_cmn *) p->access(off_cmn_) ;
//correction-anpa>>
   	hdr_cmn* ch = hdr_cmn::access(p);
//<<correction-anpa
  	hdr_pgm *ph = (hdr_pgm *) p->access(off_pgm_);

  	if ( type == PT_PGM_ODATA ) 
  		if ( acker_addr_ == 0 ) {
     		ph->options() |= PGM_OPT_PRESENT;
     		ph->opttype() |= PGM_OPT_SEND_NAK;
     		ch->size() += OPT_SEND_NAK_SIZE;
  		} else {
     		hdr_pgm_odata *poh = (hdr_pgm_odata *) p->access(off_pgm_odata_);
     		poh->acker() = acker_addr_ ;
     		ph->options() |= PGM_OPT_PRESENT;
     		ph->opttype() |= PGM_OPT_ELECT ;
     		ch->size() += OPT_ELECT_SIZE ;
  		}

  	if ( use_timestamp_ && type == PT_PGM_ODATA ) {
     	hdr_pgm_odata *poh = (hdr_pgm_odata *) p->access(off_pgm_odata_);
     	ph->options() |= PGM_OPT_PRESENT;
     	ph->opttype() |= PGM_OPT_TIMESTAMP ;
   	 	ch->size() += OPT_TIMESTAMP_SIZE ;
     	poh->timestamp() = Scheduler::instance().clock() ;
  	}

  	PGMSndAgent::buildoptions(p,type) ;
}

void
PGMCCSndAgent::print_xplot(Packet * p) 
{
  if ( !xplotchannel_ )
      return ;

//correction-anpa-   hdr_cmn * ch = (hdr_cmn *) p->access(off_cmn_) ;
//correction-anpa>>
   	hdr_cmn* ch = hdr_cmn::access(p);
//<<correction-anpa

  if ( ch->ptype() != PT_PGM_ACK ) { 
     PGMAgent::print_xplot(p) ;
     return ;
  }

  hdr_pgm * ph = (hdr_pgm *) p->access(off_pgm_) ;
  char xstr[100] ;
  sprintf(xstr, ". %f %u red\n", Scheduler::instance().clock(), ph->sqn() ) ;
  Tcl_Write(xplotchannel_, xstr, strlen(xstr));
}

// PGMCC Receiver Agent

static class PGMCCRcvAgentClass : public TclClass {
public:
  PGMCCRcvAgentClass() : TclClass("Agent/PGM/Rcv/CC") {}
  TclObject* create(int, const char*const*) {
    return (new PGMCCRcvAgent());
  }
} class_pgm_rcv_cc_agent;


PGMCCRcvAgent::PGMCCRcvAgent() :
ack_bitmask_(0),
rx_loss_(0),
rx_do_ack_(0),
xplotlosschannel_(0),
last_ts_(0),
last_rcvd_(0),
tfrc_intervals_(9)		// 8 used, 1 spare
{
    bind("weight_", &weight_) ;
    bind("tfrc_intervals_", &tfrc_intervals_) ;
    bind("slow_timo_", &slow_timo_) ;
    bind_bool("use_tfrc_", &use_tfrc_) ;
    bind_bool("use_slow_timo_", &use_slow_timo_) ;

    // TFRC initialization
    tfrc_loss_ivl_ = (int *) malloc(tfrc_intervals_ * sizeof(int)) ;
    tfrc_w_ = (double *) malloc(tfrc_intervals_ * sizeof(double)) ;
    tfrc_sum_w_ = 0 ;
    tfrc_p_ = 0 ;
	tfrc_lost_ = 0 ;
    for ( int i = 0 ; i < tfrc_intervals_ ; i++ ) {
        tfrc_loss_ivl_[i] = 0 ;
   		tfrc_w_[i] = 1 ;
        if ( i >= tfrc_intervals_/2 ) {
			double t = (double) tfrc_intervals_ / 2.0 ;
	    	tfrc_w_[i] -= ( ((double) i + 1.0) - t ) / ( t + 1.0 ) ;
		}
        tfrc_sum_w_ += tfrc_w_[i] ;
    }
}

int
PGMCCRcvAgent::command(int argc, const char*const* argv)
{       
  Tcl& tcl = Tcl::instance();

  if ( argc == 3 )
    if ( !strcmp(argv[1], "loss-report") ) {
       int mode;
       char* id = (char *) argv[2]; 
       xplotlosschannel_ = Tcl_GetChannel(tcl.interp(), id, &mode);
       if (xplotlosschannel_ == 0) {
           tcl.resultf("trace: can't attach %s for writing", id);
           return (TCL_ERROR);
       }
       return (TCL_OK);
    }
  
   return PGMRcvAgent::command(argc, argv);
}

void 
PGMCCRcvAgent::recv(Packet *p, Handler* h)
{
//correction-anpa-   hdr_cmn * ch = (hdr_cmn *) p->access(off_cmn_) ;
//correction-anpa>>
   	hdr_cmn* ch = hdr_cmn::access(p);
//<<correction-anpa
  	hdr_pgm * ph = (hdr_pgm *) p->access(off_pgm_);
  	hdr_pgm_spm * psh ;
  	u_int dp_lead, dp_seq ;
  	int delta ;
 
  	if ( ph->options() & PGM_OPT_PRESENT ) 
     	option_handler(p) ;

  	if (state_ == NEW) {
     	/*
      	 * First SPM or ODATA makes receiver socket connected.
      	 */
     	if ( ch->ptype() == PT_PGM_ODATA )
       		dp_seq = ph->sqn() ; // packet sequence number
     	else if ( ch->ptype() == PT_PGM_SPM ) {
      		hdr_pgm_spm *psh = (hdr_pgm_spm *) p->access(off_pgm_spm_) ;
       		dp_seq = psh->lead() + 1 ; // XXX
     	} else
       		goto fail ;
    
     	state_ = CONNECTED ;
     	src_nla_ = ph->tsi() ;  /* XXX next pkts should have same tsi */
     	if ( rxw_irs_ == 0 ) 
        	rxw_irs_ = rxw_trail_ = rxw_lead_ = rxw_next_ = dp_seq ;
  	} 

	switch ( ch->ptype() ) {
 	case PT_PGM_ODATA:
     	// update the bitmask 
    	dp_seq = dp_lead = ph->sqn();
     	delta = (int) (dp_lead - rxw_lead_);
     	if ( pgm_seq_gt(dp_lead, rxw_lead_) ) {
			if (delta > 31)
	    		ack_bitmask_ = 0 ;
			else
	    		ack_bitmask_ <<= delta ;
			ack_bitmask_ |= 1 ;
     	} else { /* old, out of sequence ? */
			delta = -delta ;
			if (delta < 31)
	    		ack_bitmask_ |= (1 << delta) ;
     	}
   
     	// update rx_loss_ 
     	if ( pgm_seq_gt(dp_lead, rxw_lead_) ) 
			rx_loss_ *= pow( weight_, dp_lead - rxw_lead_ );
     	else if ( !reass_q_.empty() ) {
        	// search for the packet in the reassembly queue, 
        	// if it is not duplicate, decrease the loss rate 
        	ReassemblyQueue * q = reass_q_.nak_next_ ;
        	for ( ; q != &reass_q_ ; q = q->nak_next_ ) 
          		if (q->seq() == dp_lead) {
              		rx_loss_ -= pow(weight_, rxw_lead_ - dp_lead) * (1. - weight_) ;
              		break ;
           		}
    	}
     	break;

  	case PT_PGM_SPM: 
      	psh = (hdr_pgm_spm *) p->access(off_pgm_spm_) ;
      	dp_lead = psh->lead();
      	if ( pgm_seq_leq(dp_lead, rxw_lead_) ) 
        	goto end ; 
      	rx_loss_ *= pow(weight_, dp_lead - rxw_lead_) ;
      	break;

  	default:
      	goto end;
  	}

  	for ( int i = dp_lead - 1 ; pgm_seq_gt(i , rxw_lead_) ; i-- ) 
     	rx_loss_ += pow(weight_, i - rxw_lead_) * (1. - weight_); 

  	tfrc_loss_est(dp_lead) ;

  	// am i the acker? 
  	if ( rx_do_ack_ && ch->ptype() == PT_PGM_ODATA ) 
    	output(PT_PGM_ACK, dp_lead) ;  // send the ACK

	// log loss estimate
    if ( xplotlosschannel_ ) {
        char xstr[50];
		double now = Scheduler::instance().clock() ;
        sprintf(xstr, "+ %f %f %d\n", now, rx_loss_, color_) ;
        Tcl_Write(xplotlosschannel_, xstr, strlen(xstr));
        sprintf(xstr, "x %f %f %d\n", now, tfrc_p_, color_ + 1) ;
        Tcl_Write(xplotlosschannel_, xstr, strlen(xstr));
    }   

end:
   	// pass packet up for pgm processing
   	PGMRcvAgent::recv(p,h);
   	return ;

fail:
   	Packet::free(p) ;
}

void
PGMCCRcvAgent::output(int type, u_int seq) 
{
   Packet * p ;
   hdr_ip * ih ;
   hdr_pgm * ph ;
   hdr_pgm_ack * pah ;
   hdr_pgm_nak * pnh ;
    
   switch (type) {
   case PT_PGM_ACK: 
       // allocate a new ACK
       size_ = HDR_SIZE_PGM_ACK ;
       type_ = PT_PGM_ACK ;
       p = allocpkt() ;
//correction-anpa-       ih = (hdr_ip *) p->access(off_ip_) ;
//correction-anpa>>
       ih = hdr_ip::access(p);
//<<correction-anpa
       ph = (hdr_pgm *) p->access(off_pgm_) ;
       pah = (hdr_pgm_ack *) p->access(off_pgm_ack_) ;

//correction-anpa-       ih = (hdr_ip *) p->access(off_ip_) ;
//correction-anpa>>
       ih->dst().addr_ = src_nla_;
//<<correction-anpa
       ph->tsi() = src_nla_ ;
       ph->sqn() = seq ;
       ph->options() = 0 ;
       ph->opttype() = 0 ;
       pah->bitmask() = ack_bitmask_ ;
       pah->nacker() = here_.addr_ ;
       pah->rx_loss() = use_tfrc_ ? tfrc_p_ :  rx_loss_ ;
       if ( seq > rxw_lead_ )
          pah->ack_rxw_lead() = seq;
       else
          pah->ack_rxw_lead() = rxw_lead_ ;
       pah->timestamp() = last_ts_ ;
       buildoptions(p, PT_PGM_ACK) ;
       ack_sent_++ ;
       break ;

   case PT_PGM_NAK:
       size_ = HDR_SIZE_PGM_NAK ;
       type_ = PT_PGM_NAK ;
       p = allocpkt() ;
//correction-anpa-       ih = (hdr_ip *) p->access(off_ip_) ;
//correction-anpa>>
       ih = hdr_ip::access(p);
//<<correction-anpa
       ph = (hdr_pgm *) p->access(off_pgm_) ;
       pnh = (hdr_pgm_nak *) p->access(off_pgm_nak_) ;
   
       ih->dst().addr_ = path_nla_ ;
       ph->tsi() = src_nla_ ;
       ph->sqn() = seq ;
       ph->options() = 0 ;
       ph->opttype() = 0 ;
       pnh->src_nla() = here_.addr_;
       buildoptions(p, PT_PGM_NAK) ;
       break ;

   default:
       PGMRcvAgent::output(type, seq) ;
       return ;
   }

   pkts_sent_++ ;
   send(p,0) ;
}

void
PGMCCRcvAgent::buildoptions(Packet * p, packet_t type) 
{
//correction-anpa-   hdr_cmn * ch = (hdr_cmn *) p->access(off_cmn_) ;
//correction-anpa>>
   	hdr_cmn* ch = hdr_cmn::access(p);
//<<correction-anpa
  	hdr_pgm * ph = (hdr_pgm *) p->access(off_pgm_) ;
  	hdr_pgm_nak * pnh = (hdr_pgm_nak *) p->access(off_pgm_nak_) ;

  	if ( type == PT_PGM_NAK ) {
     	ph->options() |= PGM_OPT_PRESENT ;
     	ch->size() += OPT_LOSSRATE_SIZE ; 
     	ph->opttype() |= PGM_OPT_LOSSRATE ;
     	pnh->rx_loss() = use_tfrc_ ? tfrc_p_ :  rx_loss_ ;
     	pnh->rx_lead() = rxw_lead_ ;
     	// put the last timestamp received modified by 
     	// the time passed since reception of last pkt 
		double backoff = Scheduler::instance().clock() - last_rcvd_ ;
		// round up backoff time to slow_timo_ 
		if ( use_slow_timo_ ) 
			backoff = ((int) (backoff / slow_timo_) + 1) * slow_timo_ ;
     	pnh->timestamp() = last_ts_ + backoff ;
     	pnh->nacker() = here_.addr_ ;
  	}

  	// look for other options 
  	PGMRcvAgent::buildoptions(p, type) ;
} 

void
PGMCCRcvAgent::option_handler(Packet * p) 
{
  hdr_pgm * ph = (hdr_pgm *) p->access(off_pgm_) ;
    
  if ( !ph->opttype() ) 
     return ; 

  if ( ph->opttype() & PGM_OPT_ELECT ) {
     ph->opttype() &= ~PGM_OPT_ELECT ;
     hdr_pgm_odata * poh = (hdr_pgm_odata *) p->access(off_pgm_odata_) ;
     if ( state_ != NEW ) 
        rx_do_ack_ = (poh->acker() == here_.addr_) ; // am i the acker ?
  } else 
     rx_do_ack_ = 0;

  if ( ph->opttype() & PGM_OPT_SEND_NAK ) {
     ph->opttype() &= ~PGM_OPT_SEND_NAK ;
     // XXX fake a very high sequence number
     if ( path_nla_ ) 
        output(PT_PGM_NAK, rxw_lead_ ^ 0x40000000) ;
  }

  if ( ph->opttype() & PGM_OPT_TIMESTAMP ) {
     ph->opttype() &= ~PGM_OPT_TIMESTAMP ;
     hdr_pgm_odata * poh = (hdr_pgm_odata *) p->access(off_pgm_odata_) ;
     last_ts_ = poh->timestamp() ;
     last_rcvd_ = Scheduler::instance().clock() ;
  }

  PGMRcvAgent::option_handler(p) ;
}

void
PGMCCRcvAgent::tfrc_loss_est(u_int seqno) 
{
    // XXX we dont deal with out of order packets 
    if ( pgm_seq_lt(seqno, rxw_lead_) )
        return ;

    // update loss_ivl for TFRC method
    if ( pgm_seq_gt(seqno, rxw_lead_ + 1) ) {
		tfrc_lost_ = 1 ;
        for ( int i = tfrc_intervals_ - 1 ; i > 0 ; i-- )
            tfrc_loss_ivl_[i] = tfrc_loss_ivl_[i - 1] ;
        tfrc_loss_ivl_[0] = 0 ;
    } 
    tfrc_loss_ivl_[0]++ ;

	if ( !tfrc_lost_ ) {
		tfrc_p_ = 0.0 ;
		return ;
	}

    // first, we estimate the average loss interval, without discount
    double avg = 0.0 ;
    for ( int i = 1 ; i < tfrc_intervals_ ; i++ )      
		avg += tfrc_w_[i-1] * tfrc_loss_ivl_[i] ;
    avg /= tfrc_sum_w_ ;

    // find the discount value
    double discount = ( tfrc_loss_ivl_[0] > 2 * (int) avg )? 
					  max( 0.5, 2.0 * avg / tfrc_loss_ivl_[0]) : 1.0 ;

    // compute two averages (with and without tfrc_loss_ivl_[0]) 
    double avg_new = tfrc_loss_ivl_[0] ;
    for ( int i = 1 ; i < tfrc_intervals_ - 1 ; i++ )      
		avg_new += discount * tfrc_w_[i] * tfrc_loss_ivl_[i] ;
    avg_new /= tfrc_sum_w_ ;

    avg = 0 ;
    for ( int i = 1 ; i < tfrc_intervals_ ; i++ )      
		avg += discount * tfrc_w_[i-1] * tfrc_loss_ivl_[i] ;
    avg /= tfrc_sum_w_ ;

    tfrc_p_ = 1.0 / max(avg_new, avg) ; 
}
