/*
 * pgmcc.h - 000330
 * Copyright (c) 2000 Gianluca Iannaccone, Luigi Rizzo
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *      
 */

#ifndef ns_pgmcc_h
#define ns_pgmcc_h

#include "agent.h"
#include "packet.h"
#include "my_pgm.h"
#include "my_pgmsnd.h"
#include "my_pgmrcv.h"

// PGMCC Sender Agent
class PGMCCSndAgent: public PGMSndAgent {
public:
   PGMCCSndAgent();
   virtual int command(int argc, const char*const* argv) ;
   virtual void recv(Packet *, Handler *);
   virtual void timeout(int tno);

protected:
   virtual void start() ;
   virtual int can_tx();
   virtual void next_packet();
   virtual void option_handler(Packet *) ;
   virtual void buildoptions(Packet *, packet_t) ;
   virtual void print_xplot(Packet *) ;

   Tcl_Channel xplotlosschannel_ ;
   Tcl_Channel xplotpktlosschannel_ ;

   PGMTimer cc_timer_;
   double cc_timeout_;

   double ss_thresh_;
   int tcp_ss_ ;
   int dup_thresh_;
   double weight_;

   double cc_token_; 
   double cc_window_; 
   int dupacks_; 
   u_int ack_lead_;
   u_int ignore_cong_;
   u_int32_t ack_bitmask_; 

   double acker_mrtt_;
   double acker_loss_;
   double hyst;
   nsaddr_t acker_addr_; 
   int use_timestamp_ ;
   int use_ewma_rtt_ ;
   int use_adaptive_ss_ ;
   int session_startup_ ;
} ;

// PGMCC Receiver Agent

class PGMCCRcvAgent: public PGMRcvAgent {
public:
  PGMCCRcvAgent() ;
  virtual int command(int argc, const char*const* argv) ;
  virtual void recv(Packet* p, Handler*) ;
protected:
  virtual void output(int, u_int) ;
  virtual void buildoptions(Packet *, packet_t) ;
  virtual void option_handler(Packet * p) ;

  u_int32_t ack_bitmask_ ;
  double rx_loss_ ;
  double weight_ ;
  int rx_do_ack_ ;

  Tcl_Channel xplotlosschannel_ ;

  double last_ts_ ;
  double last_rcvd_ ;
  double slow_timo_ ;
  int use_slow_timo_ ;

  // TFRC parameters 
  void tfrc_loss_est(u_int seqno) ;
  int use_tfrc_ ;
  int tfrc_intervals_ ;
  int * tfrc_loss_ivl_ ;
  double * tfrc_w_ ;
  double tfrc_sum_w_ ;
  double tfrc_p_ ;
  int tfrc_lost_ ;
} ;

#endif
