#ifndef ns_pgm_rep_h
#define ns_pgm_rep_h

#include <math.h>
#include <tcl.h>
#include <iostream.h>

#include "config.h"
#include "heap.h"
#include "agent.h"
#include "packet.h"
#include "queue.h"
#include "rng.h"
#include "replicator.h" 
#include "my_pgm.h"

#define PGM_NET_TIMER_NAK_REPEAT   10
#define PGM_NET_TIMER_WAIT_FOR_NCF 11
#define PGM_NET_TIMER_NAK_ELIM     12
#define PGM_NET_TIMER_KEEP_STATE   13
#define PGM_NET_TIMER_KEEP_RTX_STATE   14

// A list of integers

class IntList {
private:
  class IntListElement {
  public:
    IntListElement(int e, IntListElement *n) : element_(e), next_(n) { }
    int element_;
    IntListElement *next_;
  };
public:
  IntList() : head_(NULL) { }
  ~IntList() { delete_list(); }
  int empty(void) { return (head_ == NULL); }
  void push(int new_element) {
    IntListElement *new_ile = new IntListElement(new_element, head_);
    head_ = new_ile;
  }
  int pop(void) {
    if (head_ == NULL) return -1;
    int element = head_->element_;
    IntListElement *tmp = head_;
    head_ = head_->next_;
    delete tmp;
    return element;
  }
  int lookup(int e) {
    IntListElement *cur = head_;

    while (cur != NULL) {
      if (cur->element_ == e)
	return 1;
      cur = cur->next_;
    }
    return 0;
  }
  int remove(int e) {
    if (head_ == NULL) return 0;

    int count = 0;
    IntListElement *ilep = head_;
    while (ilep->next_ != NULL) {
      if (ilep->next_->element_ == e) {
	IntListElement *tmp = ilep->next_;
	ilep->next_ = tmp->next_;
	delete tmp;
	count++;
      }
      else {
	ilep = ilep->next_;
      }
    }
    if (head_->element_ == e) {
      IntListElement *tmp = head_;
      head_ = tmp->next_;
      delete tmp;
      count++;
    }
    return count;
  }
  void delete_list(void) {
    while (head_ != NULL) {
      IntListElement *tmp = head_;
      head_ = head_->next_;
      delete tmp;
    }
  }
private:
  IntListElement *head_;
};

class PGMNetTimer : public PGMTimer {
public:
  PGMNetTimer(PGMAgent *a, int t, int seq) : PGMTimer(a,t), seq_(seq) {}

protected:
  virtual void expire(Event *) { agent_->timeout(timer_type_) ; }
  int seq_;
};

// per sqn retransmit state for network elements

class PGMNetRtxInfo {
public:
  PGMNetRtxInfo(PGMAgent *a, nsaddr_t t, unsigned int s, nsaddr_t n, double loss, unsigned int lead);
  ~PGMNetRtxInfo();
  unsigned int& sqn(void) { return sqn_; }
  nsaddr_t& tsi(void) { return tsi_; }
  nsaddr_t& src_nla(void) { return src_nla_; }
  nsaddr_t& nacker(void) { return nacker_; }
  double& rx_loss(void) { return rx_loss_;}
  unsigned int& rx_lead(void) { return rx_lead_;}
  PGMNetTimer& nak_repeat_timer(void) { return nak_repeat_timer_; }
  PGMNetTimer& ncf_wait_timer(void) { return nak_wait_for_ncf_timer_; }
  PGMNetTimer& nak_elim_timer(void) { return nak_elim_timer_; }
  PGMNetTimer& keep_rtx_state_timer(void) { return keep_rtx_state_timer_; }
  void push_iface(int iface) { iface_list_.push(iface); }
  void push_agent(nsaddr_t addr) { agent_list_.push(addr); }
  int pop_iface(void) { return iface_list_.pop(); }
  nsaddr_t pop_agent(void) { return agent_list_.pop(); }
  int iface_exists(void) { return !(iface_list_.empty()); }
  int agent_exists(void) { return !(agent_list_.empty()); }
protected:
  // state
  unsigned int sqn_;
  nsaddr_t tsi_;
  nsaddr_t src_nla_;
  nsaddr_t nacker_;
  double rx_loss_;
  unsigned int rx_lead_;
  IntList iface_list_; // list of interfaces to forward RDATA for this sqn_ to
  IntList agent_list_; // list of local agents to forward RDATA for this sqn_ to
  
  // timers
  PGMNetTimer nak_repeat_timer_;
  PGMNetTimer nak_wait_for_ncf_timer_;
  PGMNetTimer nak_elim_timer_;
  PGMNetTimer keep_rtx_state_timer_;
};


// PGM Network Element Agent 

class DemuxReplicator: public Replicator {
public:
  DemuxReplicator() {}
  // all function defined in Tcl
};

class PGMNetAgent : public PGMAgent, public DemuxReplicator {

public:
  PGMNetAgent();
  virtual ~PGMNetAgent() { }
  virtual void recv(Packet* p, Handler* h);
  virtual int command(int argc, const char*const* argv);
  virtual void timeout(int tno, unsigned int sqn);

protected:

  // state
  nsaddr_t spm_path_; // upstream PGM neighbor
  nsaddr_t tsi_; // the TSI for this network element (there is a PGMNetAgent
                 // for each (S,G) pair)
  nsaddr_t pgm_group_; // the group address G for this network element
  double nak_repeat_time_;
  double nak_wait_for_ncf_time_;
  double nak_elim_time_;
  double keep_rtx_state_time_;
  Tcl_HashTable *rtxstate_;

  Packet *allocnakpkt(nsaddr_t nak_tsi, unsigned int nak_sqn, nsaddr_t src_nla, double loss, unsigned int lead) ;
				 
  PGMNetRtxInfo *rtxstate_init(nsaddr_t t, unsigned int s, nsaddr_t n, double loss, unsigned int lead) ;
					  
  PGMNetRtxInfo *rtxstate_lookup(unsigned int s);
  void rtxstate_delete(unsigned int s);

};


#endif
