/*
 * pgmrcv.h - 000330
 * Copyright (c) 2000 Gianluca Iannaccone, Luigi Rizzo
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef ns_pgmrcv_h
#define ns_pgmrcv_h

#include "packet.h"
#include "queue.h"
#include "my_pgm.h"

class PGMRcvAgent ;
class ReassemblyQueue ;

class PGMRetxTimer : public TimerHandler {
public: 
  PGMRetxTimer() {}
  PGMRetxTimer(PGMRcvAgent * a, ReassemblyQueue * r, int t) :
       agent_(a), timer_type_(t), nak_(r) {}
  void resched(double delay, int t) {
     timer_type_ = t ;
     TimerHandler::resched(delay) ;
  }
  int type() { return timer_type_ ; }

protected:
  inline void expire(Event *) ;
  PGMRcvAgent * agent_ ;
  int timer_type_ ;
  ReassemblyQueue * nak_ ;
} ;

enum reass_queue_type { T_ODATA, T_HOLE, T_NAK } ;

class ReassemblyQueue {
public:
  ReassemblyQueue() ;
  ReassemblyQueue(u_int seq, int size) ;
  ReassemblyQueue(u_int seq, double time, PGMRcvAgent * a) ;
  virtual void hole() ;
  virtual void fill(int size) ;
  virtual void link_nak(ReassemblyQueue * prev) {
    nak_next_ = prev->nak_next_ ;
    prev->nak_next_ = this ;
    nak_prev_ = prev ;
    nak_next_->nak_prev_ = this ;
  }
  virtual void unlink_nak() {
    nak_next_->nak_prev_ = nak_prev_ ;
    nak_prev_->nak_next_ = nak_next_ ;
    nak_prev_ = NULL ;
    nak_next_ = NULL ;
  }
  virtual void link(ReassemblyQueue * prev) {
    seg_next_ = prev->seg_next_ ;
    prev->seg_next_ = this ;
    seg_prev_ = prev ;
    seg_next_->seg_prev_ = this ;
  }
  virtual void unlink() {
    seg_prev_->seg_next_ = seg_next_ ;
    seg_next_->seg_prev_ = seg_prev_ ;
  }
  inline int empty() { return ( seg_next_ == this ) ; }
  inline u_int seq() { return seq_ ; }
  inline reass_queue_type type() { return type_ ; } 
  inline int pkt_size() { return (type_ == T_ODATA)? u.pkt_size_ : 0 ; }
  inline void reset_ncf_retries() { u.nak.ncf_retry_ = 0 ; }
  inline int ncf_retry(int max) { return (++u.nak.ncf_retry_ >= max) ; } 
  inline int rdata_retry(int max) { return (++u.nak.rdata_retry_ >= max) ; } 

  ReassemblyQueue * seg_next_ ;
  ReassemblyQueue * seg_prev_ ;
  ReassemblyQueue * nak_next_ ;
  ReassemblyQueue * nak_prev_ ;

  PGMRetxTimer timeout_ ;     

protected:
  u_int seq_ ;
  reass_queue_type type_ ;
  union u_tag {
     int pkt_size_ ;
     struct nak_tag {
        int ncf_retry_ ;  
        int rdata_retry_ ;
     } nak ;
  } u ;
} ;

class PGMRcvAgent : public PGMAgent {
public:
  PGMRcvAgent() ;
  virtual void recv(Packet* p, Handler* h);
  virtual void output(int type, u_int sqn) ;
  virtual void timeout(ReassemblyQueue * nak, int type);

protected:
  virtual void buildoptions(Packet *, packet_t) ;
  virtual void option_handler(Packet *) ;
  
  enum RcvrStates { NEW, CONNECTED } ;
  RcvrStates state_ ;

  ReassemblyQueue reass_q_ ;  // (looks like) reassembly queue 
  virtual void clean_reass() ;

  nsaddr_t path_nla_ ;
  nsaddr_t src_nla_ ;

  u_int rxw_trail_ ;	// oldest recoverable segment
  u_int rxw_lead_ ;	// highest seqno so far
  u_int rxw_lookahead_ ;// offset beyond rxw_lead for acceptable segments
#define PGM_DEFAULT_LOOKAHEAD		10
  u_int rxw_irs_ ;	// initial receive sequence number 
  u_int rxw_next_ ; 	// next packet to read 

  u_int recvspace_ ;	// in packets 
  u_int pkts_stored_ ;	// ODATA packets stored in the reassembly queue

  double nak_bo_ivl_ ;  // max nak backoff 
  double ncf_to_ivl_ ;	// ncf timeout
  double rdata_to_ivl_ ;// rdata timeout  
  int ncf_retries_ ;	// ncf retries
  int rdata_retries_ ;	// rdata retries 
} ;

#endif
