/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the MASH Research
 * 	Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Header: /nfs/jade/vint/CVSROOT/ns-2/apps/rtp.cc,v 1.26 2000/08/18 18:34:01 haoboy Exp $";
#endif


#include <stdlib.h>

#include "config.h"
#include "agent.h"
#include "random.h"
#include "rtp.h"
//tfmcc>>
#include <sys/types.h>
#include <math.h>
#include <assert.h>
#include <float.h>
 
#include "tfmcc.h"

#define MAX(a, b)  (((a) > (b)) ? (a) : (b))                             
#define MIN(a, b)  (((a) < (b)) ? (a) : (b)) 
//tfmcc<< 


int hdr_rtp::offset_;

class RTPHeaderClass : public PacketHeaderClass {
public: 
        RTPHeaderClass() : PacketHeaderClass("PacketHeader/RTP",sizeof(hdr_rtp)) 
	{
		bind_offset(&hdr_rtp::offset_);
	}
} class_rtphdr;



static class RTPAgentClass : public TclClass {
public:
	RTPAgentClass() : TclClass("Agent/RTP") {}
	TclObject* create(int, const char*const*) {
		return (new RTPAgent());
	}
} class_rtp_agent;


RTPAgent::RTPAgent() : Agent(PT_RTP), session_(0), lastpkttime_(-1e6), 
    running_(0), rtp_timer_(this),
//tfmcc>>
	send_timer_(this), NoFeedbacktimer_(this), oldrate_(0), maxrate_(0) {
	bind("fairSize_", &fairsize_);
	bind("ndatapack_", &ndatapack_);
	bind("InitRate_", &InitRate_);
	bind("overhead_", &overhead_);
	bind("ssmult_", &ssmult_);
	bind("printStatus_", &printStatus_);
	bind("rate_", &rate_);
	bind("bval_", &bval_);
	bind("min_rate_", &min_rate_);
	bind("t_factor_", &t_factor_);
	bind("Initial_RTT_", &Initial_RTT_);
	bind("sndr_id_", &sndr_id);
//tfmcc<<
	bind("seqno_", &seqno_);
	bind_time("interval_", &interval_);
	bind("packetSize_", &size_);
	bind("maxpkts_", &maxpkts_);
	bind("random_", &random_);
//anpa>>
        bind("sgsns_per_ggsn_", &sgsns_per_ggsn_);
	bind("rncs_per_sgsn_", &rncs_per_sgsn_);
	bind("ues_per_rnc_", &ues_per_rnc_);
//<<anpa
}

//void RTPAgent::start()
//{ 
//        running_ = 1;
//        sendpkt();
//        rtp_timer_.resched(interval_);
//}

//void RTPAgent::stop()
//{   
//        rtp_timer_.force_cancel();
//        finish();
//}

//void RTPAgent::sendmsg(int nbytes, const char* /*flags*/)
//{
//        Packet *p;
//        int n;
//
//        if (++seqno_ < maxpkts_) {
//                if (size_)
//                        n = nbytes / size_;
//                else
//                        printf("Error: RTPAgent size = 0\n");
//
//                if (nbytes == -1) {
//                        start();
//                        return;
//                }
//                while (n-- > 0) {
//                        p = allocpkt();
//                        hdr_rtp* rh = hdr_rtp::access(p);
//                        rh->seqno() = seqno_;
//                        target_->recv(p);
//                }
//                n = nbytes % size_;
//                if (n > 0) {
//                        p = allocpkt();
//                        hdr_rtp* rh = hdr_rtp::access(p);
//                        rh->seqno() = seqno_;
//                        target_->recv(p);
//                }
//                idle();
//        } else {
//                finish();
                // xxx: should we deschedule the timer here? */
//        };
//}

void RTPAgent::timeout(int)
{
	if (running_) {
		sendpkt();
		
		if (session_)
			session_->localsrc_update(size_);
		double t = interval_;
		if (random_)
			/* add some zero-mean white noise */
			t += interval_ * Random::uniform(-0.5, 0.5);
		rtp_timer_.resched(t);
	}
}

/*
 * finish() is called when we must stop (either by request or because
 * we're out of packets to send.
 */
void RTPAgent::finish()
{
        running_ = 0;
        Tcl::instance().evalf("%s done", this->name());
}

void RTPAgent::advanceby(int delta)
{
        maxpkts_ += delta;
        if (seqno_ < maxpkts_ && !running_)
                start();
}               


void RTPAgent::recv(Packet *pkt, Handler *) {
//anpa>>
// If the hole network topology has not been created yet, it must be created.
	if (net_topology.GetSize_of_list() == 0) Create_Net_Topology();
//anpa<<
	
//	hdr_rtp* rh = hdr_rtp::access(pkt);
	if (net_topology.Node_is_EXT(here_.addr_))
		receive_rtcp(pkt, 0);
	else
		m_routing(pkt);

	if (session_)
	{
		session_->recv(pkt, 0);
	}
	else
	{
		Packet::free(pkt);
	}
}

void RTPAgent::m_routing(Packet *pkt)
{
	nsaddr_t dest;
	if (net_topology.Node_is_GGSN(here_.addr_)) {
		list<int> SGSN_list = net_topology.Get_SGSNs_from_UEs(net_topology.Get_all_UEs());
		for (list<int>::iterator i = SGSN_list.begin() ; i != SGSN_list.end() ; i++) {
			dest = nsaddr_t(*i);
			fwd_rtp_pkt(pkt, dest);
		}
	} else if (net_topology.Node_is_SGSN(here_.addr_)) {
		list<int> RNC_list = net_topology.Get_RNCs_from_UEs(net_topology.Get_all_UEs(), here_.addr_);
		for (list<int>::iterator i = RNC_list.begin() ; i != RNC_list.end() ; i++) {
			dst_.addr_ = nsaddr_t(*i);
			fwd_rtp_pkt(pkt, dest);
		}
	} else if (net_topology.Node_is_RNC(here_.addr_)) {
		list<int> UE_list = net_topology.Get_UEs_under_RNC(net_topology.Get_all_UEs(), here_.addr_);
		for (list<int>::iterator i = UE_list.begin() ; i != UE_list.end() ; i++) {
			dst_.addr_ = nsaddr_t(*i);
			fwd_rtp_pkt(pkt, dest);
		}
	} else if (net_topology.Node_is_UE(here_.addr_))
		printf("RTP Agent in UE took the packet: %i\n", here_.addr_);
	  else 
		printf("This node has no identity!\n");
}

int RTPAgent::command(int argc, const char*const* argv)
{
	if (argc == 2) {
		if (strcmp(argv[1], "rate-change") == 0) {
			rate_change();
			return (TCL_OK);
		} else if (strcmp(argv[1], "start") == 0) {
                        start();
			return (TCL_OK);
                } else if (strcmp(argv[1], "stop") == 0) {
                        stop();
                        return (TCL_OK);
                }
	} else if (argc == 3) {
		if (strcmp(argv[1], "session") == 0) {
			session_ = (RTPSession*)TclObject::lookup(argv[2]);
			return (TCL_OK);
		} else if (strcmp(argv[1], "advance") == 0) {
                        int newseq = atoi(argv[2]);
                        advanceby(newseq - seqno_);
                        return (TCL_OK); 
                } else if (strcmp(argv[1], "advanceby") == 0) {
                        advanceby(atoi(argv[2]));
                        return (TCL_OK);
                }
	}
	return (Agent::command(argc, argv));
}

/* 
 * We modify the rate in this way to get a faster reaction to the a rate
 * change since a rate change from a very low rate to a very fast rate may 
 * take an undesireably long time if we have to wait for timeout at the old
 * rate before we can send at the new (faster) rate.
 */
void RTPAgent::rate_change()
{
	rtp_timer_.force_cancel();
	
	double t = lastpkttime_ + interval_;
	
	double now = Scheduler::instance().clock();
	if ( t > now)
		rtp_timer_.resched(t - now);
	else {
		sendpkt();
		rtp_timer_.resched(interval_);
	}
}

//void RTPAgent::sendpkt()
//{
//	Packet* p = allocpkt();
//	lastpkttime_ = Scheduler::instance().clock();
//	makepkt(p);
//	target_->recv(p, (Handler*)0);
//}

//anpa>>
void RTPAgent::fwd_rtp_pkt(Packet *pkt, nsaddr_t dest)
{
// Read the header of the packet.
	hdr_cmn* cmn = hdr_cmn::access(pkt);
	int packet_id = cmn->uid();
	Packet* np = allocpkt();
	cmn = hdr_cmn::access(np);
	cmn->uid() = packet_id;
	
	hdr_rtp* pkt_h = hdr_rtp::access(pkt);
	hdr_rtp* tfmcch = hdr_rtp::access(np);
	
	tfmcch->n_seqno_ = pkt_h->n_seqno_;
	tfmcch->seqno_ = pkt_h->seqno_;
	tfmcch->supp_rate = pkt_h->supp_rate;
	tfmcch->timestamp = pkt_h->timestamp;
	tfmcch->timestamp_echo = pkt_h->timestamp_echo;
	tfmcch->timestamp_offset = pkt_h->timestamp_offset;
	tfmcch->max_rtt = pkt_h->max_rtt;
	tfmcch->psize = pkt_h->psize;
	tfmcch->round_id = pkt_h->round_id;
	tfmcch->is_clr = pkt_h->is_clr;
	tfmcch->rtt_recv_id = pkt_h->rtt_recv_id;

	tfmcch->recv_id = pkt_h->recv_id;
	tfmcch->rate = pkt_h->rate;
	tfmcch->have_rtt = pkt_h->have_rtt;
	tfmcch->have_loss = pkt_h->have_loss;
	tfmcch->receiver_leave = pkt_h->receiver_leave;
	hdr_ip* iph = hdr_ip::access(np);
	iph->saddr() = net_topology.Get_EXT_id();

	dst_.addr_ = dest;
	lastpkttime_ = Scheduler::instance().clock();
	makepkt(np);
	send(np, 0);
}
//<<anpa

void RTPAgent::makepkt(Packet* p)
{
	hdr_rtp *rh = hdr_rtp::access(p);
	/* Fill in srcid_ */
	rh->srcid() = session_ ? session_->srcid() : 0;
}

//void RTPTimer::expire(Event* /*e*/) {
//        a_->timeout(0);
//}

//anpa>>
//access functions to net_topology
void RTPAgent::Create_Net_Topology()
{ net_topology = Net_Topology(sgsns_per_ggsn_, rncs_per_sgsn_, ues_per_rnc_); }

Net_Topology RTPAgent::Get_Net_Topology()
{ return net_topology; }

//<<anpa
//tfmcc>>
void RTPAgent::start() {
	double now = Scheduler::instance().clock();

	seqno_ = 0;
	n_seqno_ = 0;
	rate_ = InitRate_;
	delta_ = 0;
	oldrate_ = rate_;  
	rate_change_ = INIT_RATE;
	last_change_ = 0;
	round_begin_ = now; // not 0, otherwise "no-feedback reduction" in first round
	maxrate_ = 0;
	ndatapack_ = 0;
	active_ = 1; 
	round_id = 0;
	expected_rate = 0 ;
	inter_packet = 0;
	max_rtt_ = Initial_RTT_;
	supp_rate_ = DBL_MAX;

	rtt_recv_id = -1;
	rtt_prio = 0;
	rtt_recv_timestamp = 0;
	rtt_recv_last_feedback = 0;
	rtt_rate = 0;

	// CLR
	clr_id = -1;
	clr_timestamp = 0;
	clr_last_feedback = 0;

//anpa>>
// If the hole network topology has not been created yet, it must be created.
	if (net_topology.GetSize_of_list() == 0) Create_Net_Topology();
//anpa<<
	
	// send the first packet
	sendpkt();
	// ... at initial rate
	send_timer_.resched(size_/rate_);
	// ... and start timer so we can cut rate 
	// in half if we do not get feedback
	NoFeedbacktimer_.resched(DEC_NO_REPORT*size_/rate_); 
}

void RTPAgent::stop() {
	active_ = 0;
	send_timer_.force_cancel();
}

/*
 * Receive a status report from the receiver.
 */
void RTPAgent::receive_rtcp(Packet *pkt, Handler *) {
	double now = Scheduler::instance().clock();
	hdr_rtp *nck = hdr_rtp::access(pkt);
	int cur_prio;
	int new_clr = 0;
	double adjusted_rate;

	if (rate_change_ == INIT_RATE)
		rate_change_ = SLOW_START;

	// if we don't get ANY feedback for some time, cut rate in half
	NoFeedbacktimer_.resched(DEC_NO_REPORT * max_rtt_);
			
	// compute receiver's current RTT
	double tmp_rtt = now - nck->timestamp_echo - nck->timestamp_offset;
	if (max_rtt_ < tmp_rtt)
		max_rtt_ = tmp_rtt;

	adjusted_rate = nck->rate;
	if (nck->have_rtt == 0 && rate_change_ == CONG_AVOID) {
		// recalculate rate with correct RTT
		adjusted_rate = adjusted_rate / tmp_rtt * max_rtt_;
	}

	if (adjusted_rate < min_rate_)
		adjusted_rate = min_rate_;

	// adjust suppression rate by G_FACTOR if non-CLR (and correct feedback round)
	// before clr_id is updated so that new CLR also contributes to suppression
	if ((clr_id != nck->recv_id) && (supp_rate_ > (1-G_FACTOR) * nck->rate) &&
	    (nck->round_id == round_id))
		supp_rate_ = (1-G_FACTOR) * nck->rate;

	// if we are in slow start and we just saw a loss then come out of slow start
	if (rate_change_ == SLOW_START && nck->have_loss) {
		rate_change_ = CONG_AVOID;
		oldrate_ = rate_ = adjusted_rate;
		clr_id = nck->recv_id;
		new_clr = 1;
	}

	if (rate_change_ == SLOW_START) {
		// adjust maxrate
		if (clr_id == nck->recv_id) {
			maxrate_ = adjusted_rate;
		} else if ((nck->receiver_leave == 0) && (adjusted_rate < maxrate_ || clr_id == -1)) {
			// change CLR
			maxrate_ = adjusted_rate;
			clr_id = nck->recv_id;
			new_clr = 1;
		}
		slowstart();
	} else if (rate_change_ == CONG_AVOID) {
		if ((nck->receiver_leave == 0) && (adjusted_rate < rate_ || clr_id == -1)) {
			// change CLR
			clr_id = nck->recv_id;
			new_clr = 1;
		}
		if (clr_id == nck->recv_id) {
			if (adjusted_rate >= 0) {
				expected_rate = adjusted_rate;
				if (expected_rate > rate_) 
					increase_rate();
				else
					decrease_rate();
			}
		}
	}

	// determine which timestamp to include in data packet
	if (clr_id == nck->recv_id) {
		// remember CLR timestamp to include in packets w/o any other RTT feedback
		clr_timestamp = nck->timestamp;
		clr_last_feedback = now;
		if (!nck->have_rtt || new_clr)
			cur_prio = PRIO_FORCE;
		else
			cur_prio = PRIO_CLR;
	} else {
		if (!nck->have_rtt)
			cur_prio = PRIO_NO_RTT;
		else
			cur_prio = PRIO_RECV;
	}
	if (cur_prio > rtt_prio || (cur_prio == rtt_prio && nck->rate < rtt_rate)) {
		rtt_prio = cur_prio;
		rtt_recv_id = nck->recv_id;
		rtt_recv_timestamp = nck->timestamp;
		rtt_recv_last_feedback = now;
		rtt_rate = nck->rate; // use reported rate for suppression
	}

	if (printStatus_) {
		if (nck->receiver_leave)
			printf("s%i %f leave-group", sndr_id, now);
		else
			printf("s%i %f update", sndr_id, now);
		if (supp_rate_ == DBL_MAX) {
			printf(" [%i,i%i,m%i] %f/%f sendrate %f (supp oo, max %f, max rtt %f) round %i(%i) ",
			       nck->recv_id, rtt_recv_id, clr_id, adjusted_rate * 8.0 / 1000.0, nck->rate * 8.0 / 1000.0,
			       rate_ * 8.0 / 1000.0, maxrate_ * 8.0 / 1000.0,
			       max_rtt_, round_id, nck->round_id);
		} else {
			printf(" [%i,i%i,m%i] %f/%f sendrate %f (supp %f, max %f, max rtt %f) round %i(%i) ",
			       nck->recv_id, rtt_recv_id, clr_id, adjusted_rate * 8.0 / 1000.0, nck->rate * 8.0 / 1000.0,
			       rate_ * 8.0 / 1000.0, supp_rate_ * 8.0 / 1000.0, maxrate_ * 8.0 / 1000.0,
			       max_rtt_, round_id, nck->round_id);
		}
		printf((nck->recv_id == clr_id) ? "!\n" : "\n"); // feedback was from rep
	}
	Packet::free(pkt);
}

void RTPAgent::slowstart () {
	double now = Scheduler::instance().clock(); 

	if (rate_+SMALLFLOAT< size_/max_rtt_ ) {
		/* if this is the first report, change rate to 1 per rtt */
		/* compute delta so rate increases slowly to new value   */
		oldrate_ = rate_;
		rate_ = fairsize_/max_rtt_;
		delta_ = (rate_ - oldrate_)/(rate_*max_rtt_/size_);
		last_change_ = now;
	} else {
		/* else multiply the rate by ssmult_, and compute delta, */
		/*  so that the rate increases slowly to new value       */
		if (maxrate_ > 0) {
			if (ssmult_*rate_ < maxrate_ && now - last_change_ > max_rtt_) {
				rate_ = ssmult_*rate_; 
				delta_ = (rate_ - oldrate_)/(rate_*max_rtt_/size_);
				last_change_ = now;
			} else {
				if (rate_ > maxrate_ || now - last_change_ > max_rtt_) {
					rate_ = maxrate_; 
					delta_ = (maxrate_ - oldrate_)/(rate_*max_rtt_/size_);
					if (delta_ < 0) delta_ = 0;
					last_change_ = now; 
				}
			}
		} else {
			if (now - last_change_ > max_rtt_) {
				rate_ = ssmult_*rate_; 
				delta_ = (rate_ - oldrate_)/(rate_*max_rtt_/size_);
				last_change_ = now;
			}
		}
	}
}

void RTPAgent::increase_rate() {
	double now = Scheduler::instance().clock();
	double mult = (now - last_change_) / max_rtt_;
	if (mult > 2) mult = 2;
	
	//printf("stmp %f | %f -> ", now, rate_*8.0/1000.0);
	rate_ = MIN(expected_rate, rate_ + (fairsize_/max_rtt_)*mult);
	//printf("%f step %f (min %f)\n", expected_rate*8.0/1000.0, rate_*8.0/1000.0, fairsize_/max_rtt_*8.0/1000.0);

	if (rate_ < fairsize_/max_rtt_)
		rate_ = fairsize_/max_rtt_;

        last_change_ = now;

	// check if rate increase should impact send timer
	double next = size_/rate_;
	if (inter_packet >= 2 * next) {
 		assert(next > SMALLFLOAT);
		send_timer_.resched(next);
	}
}       

void RTPAgent::decrease_rate() 
{
	double now = Scheduler::instance().clock();
	oldrate_ = rate_ = expected_rate;
	last_change_ = now;
}

void RTPAgent::nextpkt() {
	double now = Scheduler::instance().clock();
	double xrate = -1;

	// new feedback round every T = t_factor * max_RTT, 
	//   if at least one non-CLR FBM was received (supp_rate_ < DBL_MAX)
	//   or after T = 2 * t_factor * max_RTT (only one receiver in group)
	if ((now - round_begin_ >= t_factor_ * max_rtt_) && (supp_rate_ < DBL_MAX) ||
	    (now - round_begin_ >= 2 * t_factor_ * max_rtt_)) {
		round_id = (round_id + 1) % MAX_ROUND_ID;
		round_begin_ = now;
		supp_rate_ = DBL_MAX;
		// XXX draft says only inc when max_rtt unchanged from prev round
		max_rtt_ *= MAX_RTT_DECAY;
		if (max_rtt_ < size_/rate_) 
			max_rtt_ = size_/rate_;
	}

	// time out CLR
	if (now - clr_last_feedback > max_rtt_ * DEC_NO_REPORT) {
		clr_id = -1;
	}

	sendpkt();
	
	// during slow start, we increase rate
	// slowly - by amount delta per packet 
	if ((rate_change_ == SLOW_START) && (oldrate_+SMALLFLOAT< rate_)) {
		oldrate_ = oldrate_ + delta_;
		xrate = oldrate_;
	} else {
		xrate = rate_;
	}

	//printf("%f xrate: %f\n",Scheduler::instance().clock(),xrate);

	if (xrate > SMALLFLOAT) {
		inter_packet = size_/xrate;

		// randomize between +-50% packet interval, but no more than max_rtt gap
		double max_variation = MAX(0, MIN(0.5 * inter_packet, max_rtt_ - inter_packet));
		double next = Random::uniform(inter_packet - max_variation, inter_packet + max_variation);
 		assert(next > SMALLFLOAT);
		send_timer_.resched(next);
	}
}

void RTPAgent::sendpkt() {
	if (active_) {
		double now = Scheduler::instance().clock();
		Packet* p = allocpkt();
		hdr_ip* iph = hdr_ip::access(p);
		iph->daddr() = net_topology.Get_GGSN_id();
		hdr_rtp *tfmcch = hdr_rtp::access(p);
		tfmcch->n_seqno_=n_seqno_++;
		tfmcch->seqno_ = seqno_++;
		tfmcch->timestamp = now;
		tfmcch->supp_rate = supp_rate_;
		tfmcch->psize = size_;
		tfmcch->round_id = round_id;
		tfmcch->max_rtt = max_rtt_;

		if (clr_id == rtt_recv_id || rtt_recv_id == -1) {
			tfmcch->rtt_recv_id = clr_id;
			tfmcch->timestamp_echo =  clr_timestamp;
			tfmcch->timestamp_offset = now - clr_last_feedback;
			tfmcch->is_clr = 1;
		} else {
			tfmcch->rtt_recv_id = rtt_recv_id;
			tfmcch->timestamp_echo = rtt_recv_timestamp;
			tfmcch->timestamp_offset = now - rtt_recv_last_feedback;
			tfmcch->is_clr = 0;
		}

		send(p, 0);

		/*
		if (printStatus_) {
				if (supp_rate_ == DBL_MAX) {
					printf("s%i %f data-packet [i%i,m%i] sendrate %f (supp oo, max %f, max rtt %f) round %i\n",
					       sndr_id, now, rtt_recv_id, clr_id, rate_ * 8.0 / 1000.0, maxrate_ * 8.0 / 1000.0,
					       max_rtt_, round_id);
				} else {
					printf("s%i %f data-packet [i%i,m%i] sendrate %f (supp %f, max %f, max rtt %f) round %i\n",
					       sndr_id, now, rtt_recv_id, clr_id, rate_ * 8.0 / 1000.0, supp_rate_ * 8.0 / 1000.0,
					       maxrate_ * 8.0 / 1000.0, max_rtt_, round_id);
				}
		}
		*/
		ndatapack_++;
		rtt_recv_id = -1;
		rtt_prio = 0;
	}
}

void RTPAgent::reduce_rate_on_no_feedback()
{
	rate_ *= 0.5;
	double next =  size_ / rate_;
	assert(next > SMALLFLOAT);
	send_timer_.resched(next);

	NoFeedbacktimer_.resched(DEC_NO_REPORT*max_rtt_);

	if (printStatus_)
		printf("s%i %f no-feedback reduction\n", sndr_id, Scheduler::instance().clock());
}

void RTPTimer::expire(Event *) {
  	a_->nextpkt();
}

void RTPNoFeedbackTimer::expire(Event *) {
	a_->reduce_rate_on_no_feedback();
}

//tfmcc<<
