/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the MASH Research
 * 	Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: /nfs/jade/vint/CVSROOT/ns-2/apps/rtp.h,v 1.14 2000/09/01 03:04:06 haoboy Exp $
 */

#ifndef ns_rtp_h
#define ns_rtp_h

#include "config.h"
#include "object.h"
#include "agent.h"
#include "timer-handler.h"
#include "address.h"
#include "ip.h"
//anpa>>
#include "net_topology.h"
//<<anpa
//tfmcc>>
#include "random.h"

#define SMALLFLOAT 0.0000001

/* modes of rate change */
#define INIT_RATE 0
#define SLOW_START 1
#define CONG_AVOID 2

/* cut sending rate after DEC_NO_REPORT rtts without receiver report */
#define DEC_NO_REPORT 10
#define G_FACTOR 0.1
#define MAX_ROUND_ID 1000000 // just for logging, in real world 16 is enough
#define MAX_RTT_DECAY 0.9
//tfmcc<<

#define RTP_M 0x0080 // marker for significant events
//anpa>>
#define rtcp 1
//anpa<<

//tfmcc>>
//enum {MY_PRIO_FORCE = 4, MY_PRIO_NO_RTT = 3, MY_PRIO_RECV = 2, MY_PRIO_CLR = 1};
//tfmcc<<
/* rtp packet.  For now, just have srcid + seqno. */
struct hdr_rtp { 
	u_int32_t srcid_;
	int seqno_;
//anpa>>
	int n_seqno_;	
//anpa<<
	
	//rtp flags indicating significant event(begining of talkspurt)
	u_int16_t flags_; 
	
	static int offset_;
	inline static int& offset() { return offset_; }
	inline static hdr_rtp* access(const Packet* p) {
		return (hdr_rtp*) p->access(offset_);
	}
	
	/* per-field member functions */
	u_int32_t& srcid() { return (srcid_); }
	int& seqno() { return (seqno_); }
	int& n_seqno() { return (n_seqno_); }
	u_int16_t& flags() { return (flags_); }
//tfmcc>>
	double supp_rate;	 // suppression rate
	double timestamp; 	 // time this message was sent
	double timestamp_echo; 	 // echo timestamp from last receiver report
	double timestamp_offset; // offset since sender received report
	double max_rtt;          // maximum RTT of all receivers
	int psize;		 // packet size
	int round_id; 		 // round id.
	int is_clr;              // current representative receiver
	int rtt_recv_id;         // id of receiver whose timestamp is sent back

	inline static hdr_rtp* access(Packet* p) {
		return (hdr_rtp*) p->access(offset_);
	}

	int recv_id;                    // ID of receiver giving feedback
	double rate;                    // expected sending rate (eqn)
	int have_rtt;                   // receiver has at least one RTT measurement
	int have_loss;                  // receiver exprienced at least one loss event
	int receiver_leave;             // receiver will leave session
//tfmcc<<
};

class RTPSource : public TclObject {
public:
	RTPSource* next;

	RTPSource(u_int32_t srcid);
	inline u_int32_t srcid() { return (srcid_); }
	inline int np() { return (np_); }
	inline int snp() { return (snp_); }
	inline int ehsr() { return (ehsr_); }

	inline void np(int n) { np_ += n; }
	inline void snp(int n) { snp_ = n; }
	inline void ehsr(int n) { ehsr_ = n; }
protected:
	u_int32_t srcid_;
	int np_;
	int snp_;
	int ehsr_;
};

class RTPSession : public NsObject {
public:
	RTPSession();
	~RTPSession();
	virtual void recv(Packet* p, Handler*);
	virtual void recv_ctrl(Packet* p);
	int command(int argc, const char*const* argv);
	inline u_int32_t srcid() { return (localsrc_->srcid()); }
	int build_report(int bye);
	void localsrc_update(int);
protected:
	RTPSource* allsrcs_;
	RTPSource* localsrc_;
	int build_sdes();
	int build_bye();
	RTPSource* lookup(u_int32_t);
	void enter(RTPSource*);
	int last_np_;
};

class RTPAgent;

class RTPTimer : public TimerHandler {
public: 
        RTPTimer(RTPAgent *a) : TimerHandler() { a_ = a; }
        virtual void expire(Event *e);
protected:
        RTPAgent *a_;
};

//tfmcc>>
class RTPNoFeedbackTimer : public TimerHandler {
public:
	RTPNoFeedbackTimer(RTPAgent *a) : TimerHandler() { a_ = a; }
	virtual void expire(Event *e);
protected:
	RTPAgent *a_;
}; 
//tfmcc<<

class RTPAgent : public Agent {
//tfmcc>>
//	friend class RTPTimer;
//	friend class RTPNoFeedbackTimer;
//tfmcc<<
 public:
        RTPAgent();
        virtual void timeout(int);
        virtual void recv(Packet* p, Handler*);
        virtual int command(int argc, const char*const* argv);
        void advanceby(int delta);
//        virtual void sendmsg(int nbytes, const char *flags = 0);
//tfmcc>>
	void sendpkt();
	void nextpkt();
	void start();
	void stop();
	void increase_rate();
	void decrease_rate();
	void slowstart();
	void reduce_rate_on_no_feedback();
//tfmcc<<
protected:

	virtual void makepkt(Packet*);
        void rate_change();
        virtual void finish();
//anpa>>
	void m_routing(Packet *);
	void fwd_rtp_pkt(Packet *, nsaddr_t);
	void receive_rtcp(Packet*, Handler*);
	void Create_Net_Topology();
	Net_Topology Get_Net_Topology();

//<<anpa
//tfmcc>>
	RTPTimer send_timer_;
	RTPNoFeedbackTimer NoFeedbacktimer_;
	int seqno_;
//anpa>>
	int n_seqno_;
//anpa<<
	int psize_;             // packet size of TFMCC flow
	int fairsize_;          // packet size of competing flow
	double rate_;		// send rate
	double oldrate_;	// allows rate to be changed gradually
	double delta_;		// allows rate to be changed gradually
	int rate_change_; 	// slow start, cong avoid, decrease ...
	double expected_rate; 	// TCP friendly rate based on current RTT 
				//  and recever-provded loss estimate
	double maxrate_;	// maximum rate during slowstart (prevents
	                        //  sending at more than 2 times the 
				//  rate at which the receiver is _receiving_)
 	int printStatus_;	// to print status reports
	int sndr_id;            // only needed for status reports

	double inter_packet;    // inter packet gap

	double max_rtt_;        // (assumed) maximum RTT of all receivers
	double Initial_RTT_;    // safe initial value for max RTT
	double t_factor_;       // T = t_factor_ * max_rtt_
	double round_begin_;    // time when fb round started

	int InitRate_;		// initial send rate
	double last_change_;	// time last change in rate was made
	double ssmult_;		// during slow start, increase rate by this 
				//  factor every rtt
	int bval_;		// value of B for the formula
	double overhead_;	// if > 0, dither outgoing packets 
	TracedInt ndatapack_;	// number of packets sent
	int active_;		// have we shut down? 
	int round_id;		// round id

	double min_rate_;
	double supp_rate_;

	// extensions for multicast
	int    rtt_recv_id;
	int    rtt_prio;
	double rtt_recv_timestamp;
	double rtt_recv_last_feedback;
	double rtt_rate;

	// CLR
	int    clr_id;
	double clr_timestamp;
	double clr_last_feedback;
//tfmcc<<

	RTPSession* session_;
        double lastpkttime_;

//anpa>>
	int sgsns_per_ggsn_;
	int rncs_per_sgsn_;
	int ues_per_rnc_;
	
	Net_Topology net_topology;
//<<anpa
	int running_;
        int random_;
        int maxpkts_;
        double interval_;
        RTPTimer rtp_timer_;
};

#endif
