global ns

set ns [new Simulator]

set f [open test.tr w]
$ns trace-all $f

puts " Simulation started."

# Declare "finish" procedure which closes the trace file
proc finish {} {
	global ns f
	$ns flush-trace
	close $f
	puts " Simulation ended."
	exit 0
}


########### DECLARATIONS ###########
# Declare SGSN per GGSN ratio.
set sgsns_per_ggsn 1

# Declare RNC per GGSN ratio.
set rncs_per_sgsn 1

# Declare UE per RNC ratio.
set ues_per_rnc 2

# Declare packet size of rtp traffic
set psize 512


########### NODE CREATION ###########
# Calculate the number of nodes.
set num_of_rncs [expr {$sgsns_per_ggsn * $rncs_per_sgsn}]
set num_of_ues [expr {$num_of_rncs * $ues_per_rnc}]

# Create RNC's.
$ns node-config -UmtsNodeType rnc
for {set i 0} {$i < $num_of_rncs} {incr i} {
	set rnc($i) [$ns create-Umtsnode]
}

# Create Node Bs and Iub interfaces.
$ns node-config -UmtsNodeType bs \
		-downlinkBW 32kbs \
		-downlinkTTI 10ms \
		-uplinkBW 32kbs \
		-uplinkTTI 10ms 

for {set i 0} {$i < $num_of_rncs} {incr i} {
	set bs($i) [$ns create-Umtsnode]
	$ns setup-Iub $bs($i) $rnc($i) 622Mbit 622Mbit 15ms 15ms DummyDropTail 2000
}

# Create UE's
set j 0
set k 0
for {set i 0} {$i < $num_of_ues} {incr i} {
	if {$j == $ues_per_rnc} then {
		set j 0
	}
	if {$j == 0} then {
		$ns node-config -UmtsNodeType ue \
			-baseStation $bs($k) \
			-radioNetworkController $rnc($k)
		incr k
	}
	
	incr j
	set ue($i) [$ns create-Umtsnode]
}

# Create SGSN's.
set k 0
for {set i 0} {$i < $sgsns_per_ggsn} {incr i} {
	set sgsn($i) [$ns node]
	for {set j 0} {$j < $rncs_per_sgsn} {incr j} {
		$ns duplex-link $rnc($k) $sgsn($i) 622Mbit 0.4ms DropTail 1000
		incr k
	}
}

# Create GGSN and Gn interfaces.
set ggsn [$ns node]
for {set i 0} {$i < $sgsns_per_ggsn} {incr i} {
	$ns duplex-link $sgsn($i) $ggsn 622MBit 10ms DropTail 1000
}

# Create external node and link it to GGSN.
set ext_node [$ns node]
$ns duplex-link $ggsn $ext_node 10MBit 35ms DropTail 1000


######### AGENT CREATION #########
# Create RTP/RTCP agent in external node.
set rtp_ext [new Agent/RTP]
$rtp_ext set fid_ 0
$rtp_ext set packetSize_ $psize
$rtp_ext set sgsns_per_ggsn_ $sgsns_per_ggsn
$rtp_ext set rncs_per_sgsn_ $rncs_per_sgsn
$rtp_ext set ues_per_rnc_ $ues_per_rnc

$ns attach-agent $ext_node $rtp_ext

# Create RTP agent in GGSN.
set rtp_ggsn [new Agent/RTP]
$rtp_ggsn set fid_ 0
$rtp_ggsn set packetSize_ $psize
$rtp_ggsn set sgsns_per_ggsn_ $sgsns_per_ggsn
$rtp_ggsn set rncs_per_sgsn_ $rncs_per_sgsn
$rtp_ggsn set ues_per_rnc_ $ues_per_rnc

$ns attach-agent $ggsn $rtp_ggsn
$ns connect $rtp_ext $rtp_ggsn

# Create RTP agents in SGSNs.
for {set i 0} {$i < $sgsns_per_ggsn} {incr i} {
	set rtp_sgsn($i) [new Agent/RTP]
	$rtp_sgsn($i) set fid_ 0
	$rtp_sgsn($i) set packetSize_ $psize
	$rtp_sgsn($i) set sgsns_per_ggsn_ $sgsns_per_ggsn
	$rtp_sgsn($i) set rncs_per_sgsn_ $rncs_per_sgsn
	$rtp_sgsn($i) set ues_per_rnc_ $ues_per_rnc
	
	$ns attach-agent $sgsn($i) $rtp_sgsn($i)
	$ns connect $rtp_ggsn $rtp_sgsn($i)
}

# Create RTP agents in RNCs.
set k 0;
set rtp_rnc($k) [new Agent/RTP]
for {set i 0} {$i < $sgsns_per_ggsn} {incr i} {
	for {set j 0} {$j < $rncs_per_sgsn} {incr j} {
		set rtp_rnc($k) [new Agent/RTP]
		$rtp_rnc($k) set fid_ 0
		$rtp_rnc($k) set packetSize_ $psize
		$rtp_rnc($k) set sgsns_per_ggsn_ $sgsns_per_ggsn
		$rtp_rnc($k) set rncs_per_sgsn_ $rncs_per_sgsn
		$rtp_rnc($k) set ues_per_rnc_ $ues_per_rnc
		
		$ns attach-agent $rnc($k) $rtp_rnc($k)
		$ns connect $rtp_sgsn($i) $rtp_rnc($k)

		incr k;
	}
}

# Create RTCP agents in UEs.
set k 0;
for {set i 0} {$i < $num_of_rncs} {incr i} {
	for {set j 0} {$j < $ues_per_rnc} {incr j} {
		set rtcp_ue($k) [new Agent/RTCP]
		$rtcp_ue($k) set sgsns_per_ggsn_ $sgsns_per_ggsn
		$rtcp_ue($k) set rncs_per_sgsn_ $rncs_per_sgsn
		$rtcp_ue($k) set ues_per_rnc_ $ues_per_rnc

		$ns attach-agent $ue($k) $rtcp_ue($k)
		$ns connect $rtp_rnc($i) $rtcp_ue($k)

		incr k;
	}
}


########## TRAFFIC CREATION ##########
# Declare the video trace file and the traffic:
set trace_file64 [new Tracefile]
$trace_file64 filename video64.dat

set video64 [new Application/Traffic/Trace]
$video64 attach-tracefile $trace_file64
$video64 attach-agent $rtp_ext

# Declare the session.
set session [new Session/RTP]
$rtp_ext session $session
$rtp_ggsn session $session

# Declare RLC channels.
$ns node-config -llType UMTS/RLC/AM \
		-downlinkBW 384kbs \
		-uplinkBW 64kbs \
		-downlinkTTI 20ms \
		-uplinkTTI 20ms 

#for {set i 0} {$i < $num_of_ues} {incr i} {
#	$ns attach-common $ue($i) $rtcp_ue($i)
#}

for {set i 0} {$i < $num_of_ues} {incr i} {
	set dch($i) [$ns create-dch $ue($i) $rtcp_ue($i)]
}

# Trace UEs.
$ue(0) trace-inlink $f 2
$ue(0) trace-outlink $f 2

# Set the start/stop timing.
$ns at 0.0 "$video64 start"
$ns at 9.0 "$video64 stop"
$ns at 9.9 "finish"
$ns run